"""
$description Russian live-streaming and video hosting social platform.
$url vk.com
$type live, vod
"""

import logging
import re
from hashlib import md5
from urllib.parse import parse_qsl, unquote, urlparse

from streamlink.exceptions import NoStreamsError
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_qsd

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:\w+\.)?vk\.com/videos?(?:\?z=video)?(?P<video_id>-?\d+_\d+)"
))
@pluginmatcher(re.compile(
    r"https?://(\w+\.)?vk\.com/.+"
))
class VK(Plugin):
    API_URL = "https://vk.com/al_video.php"
    HASH_COOKIE = "hash429"

    def _get_cookies(self):
        def on_response(res, **kwargs):
            if res.headers.get("x-waf-redirect") == "1":
                if not res.headers.get("X-WAF-Backend-Status"):
                    log.debug("Getting WAF cookie")
                    cookie = res.cookies.get(self.HASH_COOKIE)
                    key = md5(cookie.encode("utf-8")).hexdigest()
                    res.headers["Location"] = update_qsd(res.headers["Location"], qsd={"key": key})
                    return res
                elif res.headers.get("X-WAF-Backend-Status") == "challenge_success":
                    self.session.http.cookies.update(res.cookies)
                    return res

        self.session.http.get("https://vk.com/", hooks={"response": on_response})

    def _has_video_id(self):
        return any(self.matches[:-1])

    def follow_vk_redirect(self):
        if self._has_video_id():
            return

        try:
            parsed_url = urlparse(self.url)
            true_path = next(unquote(v).split("/")[0] for k, v in parse_qsl(parsed_url.query) if k == "z" and len(v) > 0)
            self.url = f"{parsed_url.scheme}://{parsed_url.netloc}/{true_path}"
            if self._has_video_id():
                return
        except StopIteration:
            pass

        try:
            self.url = self.session.http.get(self.url, schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//head/meta[@property='og:url'][@content]/@content"),
                str
            ))
        except PluginError:
            pass
        if self._has_video_id():
            return

        raise NoStreamsError

    def _get_streams(self):
        self._get_cookies()
        self.follow_vk_redirect()

        video_id = self.match.group("video_id")
        if not video_id:
            return

        log.debug(f"Video ID: {video_id}")
        try:
            data = self.session.http.post(
                self.API_URL,
                params={"act": "show"},
                data={"act": "show", "al": "1", "video": video_id},
                headers={"Referer": self.url},
                schema=validate.Schema(
                    validate.transform(lambda text: re.sub(r"^\s*<!--\s*", "", text)),
                    validate.parse_json(),
                    {"payload": list},
                    validate.get(("payload", -1)),
                    list,
                    validate.get(-1),
                    {"player": {"params": [dict]}},
                    validate.get(("player", "params", 0)),
                    {
                        validate.optional("hls"): validate.url(),
                        validate.optional("manifest"): validate.startswith("<?xml"),
                        validate.optional("md_author"): validate.any(str, None),
                        validate.optional("md_title"): validate.any(str, None),
                    }
                )
            )
        except PluginError:
            log.error("Could not parse API response")
            return

        self.id = video_id
        self.author = data.get("md_author")
        self.title = data.get("md_title")

        hls = data.get("hls")
        if hls:
            return HLSStream.parse_variant_playlist(self.session, hls)

        dash_manifest = data.get("manifest")
        if dash_manifest:
            return DASHStream.parse_manifest(self.session, dash_manifest)


__plugin__ = VK



"""
$description Spanish live TV sports channel owned by Gol Network.
$url goltelevision.com
$type live
$region Spain
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?goltelevision\.com/en-directo"
))
class GOLTelevision(Plugin):
    def _get_streams(self):
        self.session.http.headers.update({
            "Origin": "https://goltelevision.com",
            "Referer": "https://goltelevision.com/",
        })
        url = self.session.http.get(
            "https://play.goltelevision.com/api/stream/live",
            schema=validate.Schema(
                validate.parse_json(),
                {"manifest": validate.url()},
                validate.get("manifest"),
            )
        )
        return HLSStream.parse_variant_playlist(self.session, url)


__plugin__ = GOLTelevision



"""
$description French free-to-air news channel, providing 24-hour national and global news coverage.
$url cnews.fr
$type live, vod
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate


@pluginmatcher(re.compile(
    r"https?://(?:www\.)?cnews\.fr"
))
class CNEWS(Plugin):
    _dailymotion_url = "https://www.dailymotion.com/embed/video/{}"

    def _get_streams(self):
        data = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"jQuery\.extend\(Drupal\.settings, ({.*})\);"),
            validate.none_or_all(
                validate.get(1),
                validate.parse_json(),
                {
                    validate.optional("dm_player_live_dailymotion"): {
                        validate.optional("video_id"): str,
                    },
                    validate.optional("dm_player_node_dailymotion"): {
                        validate.optional("video_id"): str,
                    },
                },
                validate.union_get("dm_player_live_dailymotion", "dm_player_node_dailymotion")
            ),
        ))
        if not data:
            return

        live, node = data

        if node:
            return self.session.streams(self._dailymotion_url.format(node["video_id"]))
        elif live:
            return self.session.streams(self._dailymotion_url.format(live["video_id"]))


__plugin__ = CNEWS



from streamlink.plugins.nicolive import NicoLive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlNicoLive(PluginCanHandleUrl):
    __plugin__ = NicoLive

    should_match = [
        'https://live2.nicovideo.jp/watch/lv534562961',
        'http://live2.nicovideo.jp/watch/lv534562961',
        'https://live.nicovideo.jp/watch/lv534562961',
        'https://live2.nicovideo.jp/watch/lv534562961?ref=rtrec&zroute=recent',
        'https://live.nicovideo.jp/watch/co2467009?ref=community',
        'https://live.nicovideo.jp/watch/co2619719',
    ]



"""
$description Global live-streaming platform owned by BitTorrent, Inc.
$url dlive.tv
$type live, vod
"""

import logging
import re
from urllib.parse import unquote_plus

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""
    https?://(?:www\.)?dlive\.tv/
    (?:
        p/(?P<video>[^/]+)
        |
        (?P<channel>[^/]+)
    )
""", re.VERBOSE))
class DLive(Plugin):
    URL_LIVE = "https://live.prd.dlive.tv/hls/live/{username}.m3u8"

    QUALITY_WEIGHTS = {
        "src": 1080,
    }

    @classmethod
    def stream_weight(cls, key):
        weight = cls.QUALITY_WEIGHTS.get(key)
        if weight:
            return weight, "dlive"

        return super().stream_weight(key)

    def _get_streams_video(self, video):
        log.debug(f"Getting video HLS streams for {video}")
        hls_url = self.session.http.get(self.url, schema=validate.Schema(
            validate.regex(re.compile(r'"playbackUrl"\s*:\s*"([^"]+\.m3u8)"')),
            validate.get(1),
            validate.transform(unquote_plus),
            validate.transform(lambda url: bytes(url, "utf-8").decode("unicode_escape")),
            validate.url(),
        ))

        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def _get_streams_live(self, channel):
        log.debug(f"Getting live HLS streams for {channel}")
        query = f"""query {{
            userByDisplayName(displayname:"{channel}") {{
                livestream {{
                    title
                }}
                username
            }}
        }}"""
        livestream, username = self.session.http.post(
            "https://graphigo.prd.dlive.tv/",
            json={"query": query},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "data": {
                        "userByDisplayName": {
                            "livestream": {
                                "title": str,
                            },
                            "username": str,
                        },
                    },
                },
                validate.get(("data", "userByDisplayName")),
                validate.union_get("livestream", "username"),
            ),
        )

        self.author = channel
        self.title = livestream["title"]

        return HLSStream.parse_variant_playlist(self.session, self.URL_LIVE.format(username=username))

    def _get_streams(self):
        video = self.match.group("video")
        channel = self.match.group("channel")

        if video:
            return self._get_streams_video(video)
        elif channel:
            return self._get_streams_live(channel)


__plugin__ = DLive



from streamlink.plugins.linelive import LineLive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlLineLive(PluginCanHandleUrl):
    __plugin__ = LineLive

    should_match_groups = [
        ("https://live.line.me/channels/123/broadcast/456", {"channel": "123", "broadcast": "456"}),
    ]

    should_not_match = [
        "https://live.line.me/",
        "https://live.line.me/channels/123/",
        "https://live.line.me/channels/123/upcoming/456",
    ]



"""
$description Chinese, global live-streaming platform run by Huya Live.
$url nimo.tv
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r'https?://(?:www\.|m\.)?nimo\.tv/(?P<username>.*)'
))
class NimoTV(Plugin):
    data_url = 'https://m.nimo.tv/{0}'

    video_qualities = {
        250: '240p',
        500: '360p',
        1000: '480p',
        2500: '720p',
        6000: '1080p',
    }

    _re_appid = re.compile(br'appid=(\d+)')
    _re_domain = re.compile(br'(https?:\/\/[A-Za-z]{2,3}.hls[A-Za-z\.\/]+)(?:V|&)')
    _re_id = re.compile(br'id=([^|\\]+)')
    _re_tp = re.compile(br'tp=(\d+)')

    def _get_streams(self):
        username = self.match.group('username')
        if not username:
            return

        data = self.session.http.get(
            self.data_url.format(username),
            headers={
                "User-Agent": useragents.ANDROID,
            },
            schema=validate.Schema(
                re.compile(r"<script>var G_roomBaseInfo = ({.*?});</script>"),
                validate.none_or_all(
                    validate.get(1),
                    validate.parse_json(),
                    {
                        "title": str,
                        "nickname": str,
                        "game": str,
                        "liveStreamStatus": int,
                        validate.optional("mStreamPkg"): str,
                    },
                ),
            ),
        )

        if data['liveStreamStatus'] == 0:
            log.info('This stream is currently offline')
            return

        mStreamPkg = data.get('mStreamPkg')
        if not mStreamPkg:
            log.debug('missing mStreamPkg')
            return

        mStreamPkg = bytes.fromhex(mStreamPkg)
        try:
            _appid = self._re_appid.search(mStreamPkg).group(1).decode('utf-8')
            _domain = self._re_domain.search(mStreamPkg).group(1).decode('utf-8')
            _id = self._re_id.search(mStreamPkg).group(1).decode('utf-8')
            _tp = self._re_tp.search(mStreamPkg).group(1).decode('utf-8')
        except AttributeError:
            log.error('invalid mStreamPkg')
            return

        params = {
            'appid': _appid,
            'id': _id,
            'tp': _tp,
            'u': '0',
            't': '100',
            'needwm': 1,
        }
        url = f'{_domain}{_id}.m3u8'
        log.debug(f'URL={url}')
        for k, v in self.video_qualities.items():
            _params = params.copy()
            _params["ratio"] = k
            if v == '1080p':
                _params["needwm"] = 0
            elif v in ('720p', '480p', '360p'):
                _params["sphd"] = 1

            log.trace(f'{v} params={_params!r}')
            # some qualities might not exist, but it will select a different lower quality
            yield v, HLSStream(self.session, url, params=_params)

        self.author = data['nickname']
        self.category = data['game']
        self.title = data['title']


__plugin__ = NimoTV



"""
$description French live TV channels from TF1 Group, including LCI and TF1.
$url tf1.fr
$url tf1info.fr
$url lci.fr
$type live
$region France
"""

import logging
import re

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"""
    https?://(?:www\.)?
    (?:
        tf1\.fr/(?:
            (?P<live>[\w-]+)/direct/?
            |
            stream/(?P<stream>[\w-]+)
        )
        |
        (?P<lci>tf1info|lci)\.fr/direct/?
    )
""", re.VERBOSE))
class TF1(Plugin):
    _URL_API = "https://mediainfo.tf1.fr/mediainfocombo/{channel_id}"

    def _get_channel(self):
        if self.match["live"]:
            channel = self.match["live"]
            channel_id = f"L_{channel.upper()}"
        elif self.match["lci"]:
            channel = "LCI"
            channel_id = "L_LCI"
        elif self.match["stream"]:
            channel = self.match["stream"]
            channel_id = f"L_FAST_v2l-{channel}"
        else:  # pragma: no cover
            raise PluginError("Invalid channel")

        return channel, channel_id

    def _api_call(self, channel_id):
        return self.session.http.get(
            self._URL_API.format(channel_id=channel_id),
            params={
                "context": "MYTF1",
                "pver": "4001000",
            },
            headers={
                # forces HLS streams
                "User-Agent": useragents.IPHONE,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "delivery": validate.any(
                        validate.all(
                            {
                                "code": 200,
                                "format": "hls",
                                "url": validate.url(),
                            },
                            validate.union_get("code", "url"),
                        ),
                        validate.all(
                            {
                                "code": int,
                                "error": str,
                            },
                            validate.union_get("code", "error"),
                        ),
                    ),
                },
                validate.get("delivery"),
            ),
        )

    def _get_streams(self):
        channel, channel_id = self._get_channel()
        log.debug(f"Found channel {channel} ({channel_id})")

        code, data = self._api_call(channel_id)
        if code != 200:
            log.error(data)
            return

        return HLSStream.parse_variant_playlist(self.session, data)


__plugin__ = TF1



from streamlink.plugins.oneplusone import OnePlusOne
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlOnePlusOne(PluginCanHandleUrl):
    __plugin__ = OnePlusOne

    should_match = [
        "https://1plus1.video/ru/tvguide/plusplus/online",
        "https://1plus1.video/tvguide/1plus1/online",
        "https://1plus1.video/tvguide/2plus2/online",
        "https://1plus1.video/tvguide/bigudi/online",
        "https://1plus1.video/tvguide/plusplus/online",
        "https://1plus1.video/tvguide/sport/online",
        "https://1plus1.video/tvguide/tet/online",
        "https://1plus1.video/tvguide/uniantv/online",
    ]

    should_not_match = [
        "https://1plus1.video/",
    ]



"""
$description Video content from Telefe, an Argentine TV station.
$url mitelefe.com
$type live
$region Argentina
"""

import logging
import re
from urllib.parse import urljoin

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"https://mitelefe\.com/vivo"))
class Telefe(Plugin):
    def _get_streams(self):
        self.title, hls_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(), 'HLS')]/text()"),
                validate.none_or_all(
                    re.compile(r"=\s*(\{.+?});", re.DOTALL | re.MULTILINE),
                    validate.none_or_all(
                        validate.get(1),
                        validate.parse_json(),
                        {str: {"children": {"top": {"model": {"videos": [{
                            "title": str,
                            "sources": validate.all(
                                [{"url": str, "type": str}],
                                validate.filter(lambda p: p["type"].lower() == "hls"),
                                validate.get((0, "url")),
                            ),
                        }]}}}}},
                        validate.transform(lambda k: next(iter(k.values()))),
                        validate.get(("children", "top", "model", "videos", 0)),
                        validate.union_get("title", "sources"),
                    ),
                ),
            ),
        )
        return HLSStream.parse_variant_playlist(self.session, urljoin(self.url, hls_url))


__plugin__ = Telefe



from streamlink.plugins.brightcove import Brightcove
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBrightcove(PluginCanHandleUrl):
    __plugin__ = Brightcove

    should_match = [
        "https://players.brightcove.net/123/default_default/index.html?videoId=456",
        "https://players.brightcove.net/456/default_default/index.html?videoId=789",
    ]



from streamlink.plugins.rtve import Rtve, ZTNR
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRtve(PluginCanHandleUrl):
    __plugin__ = Rtve

    should_match = [
        "https://www.rtve.es/play/videos/directo/la-1/",
        "https://www.rtve.es/play/videos/directo/canales-lineales/24h/",
        "https://www.rtve.es/play/videos/informe-semanal/la-semilla-de-la-guerra/6670279/",
    ]

    should_not_match = [
        "https://www.rtve.es",
        "http://www.rtve.es/directo/la-1",
        "http://www.rtve.es/directo/la-2/",
        "http://www.rtve.es/directo/teledeporte/",
        "http://www.rtve.es/directo/canal-24h/",
        "http://www.rtve.es/infantil/directo/",
    ]


def test_translate_no_content():
    assert list(ZTNR.translate("")) == []


def test_translate_no_streams():
    # real payload without any tEXt chunks that match the expected format
    data = \
        "iVBORw0KGgoAAAANSUhEUgAAAsAAAAGMAQMAAADuk4YmAAAAA1BMVEX///+nxBvIAAAAAXRSTlMA" \
        "QObYZgAAADlJREFUeF7twDEBAAAAwiD7p7bGDlgYAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA" \
        "AAAAAAAAwAGJrAABgPqdWQAAAcp0RVh0ak9lNmRyNkUtV2hmeEE0dERMdS9FOTlCT2d3MF9HMDdG" \
        "RmxQNy1ZLTdFOFRac0MmbD93VEp5SENvUUlseVY1bjdrYmF2ZkhUUjc4aTBHAEBxY08zdk4yYldE" \
        "bm09TDVaNGMyVVpNdklVbS5LVUNCUTdZNVpfSUZMVmRMNlN0VE14TmFPLUFGaTF6ai9YenE6PVg9" \
        "dnJBb3BFU3BBJlpoWFViSER3MCZxbj9AS0d1Si5OSnAudiMwMTYxNDA2MDU2NjcyMDE3MzI4NTcw" \
        "ODcwMDc3MDI3NjIwNjczMTA0ODEyNDY3MzMwNzgxMDQwMTE4NzQ4MDYwMjIwODgxNTI0ODEzNjQ4" \
        "MjU0MTEyMzEyNjUxMzc2NTM3MTMzNzgwNTYwNDE0NjI4NDM1NjIzNTA1MTAxNjYwMDExNzE4MDQx" \
        "MTc3MDMxNTQ2MDEzNDUwMDQ2MTg4MDgwNzMxNDM3MjgwMDQ4NDA3Mzg0MzYxODA0NjU0NDYzMTY1" \
        "NDIxMzY4ODAzNTQ3MjMyMjYzODUwMzY5MTE3MTMwOTMzMjAwNDg1MDExNTE4MTgxMTgwMTAwNjU0" \
        "NTg1MzcxNDQ5MDM5MzY2ODMxNTc0MjUyNDVZsdrfAAAAAElFTkSuQmCC"
    assert list(ZTNR.translate(data)) == []


def test_translate_has_streams():
    # real payload with modified end (IEND chunk of size 0), to reduce test size
    data = \
        "iVBORw0KGgoAAAANSUhEUgAAAVQAAAFUCAIAAAD08FPiAAACr3RFWHRXczlVSWdtM2ZPTGY4b2R4" \
        "dWo5aHZnRlRhOndvZEtxN3pLOG5oNGRpbT1vREBTWHhOMGtzUVomNndAWkV5cz1GOUlCSiYxdDcy" \
        "QmdDOFM2NGFVJmh1Nzk2bUpwOFVJOE1DJlpAY2lzdGcmbEUmRE5DZFV4SHpEOFgvLmppZ1l4b3M1" \
        "QU1lOnl3ZS04VlBwQkZvLlFMUWZHTy1vQjNVeHhfVDF1JkRSQTpPP2J4Wm0zbFlxS3IjAEhEX1JF" \
        "QURZJSUwNTYwNzI4Mjg4MzUyNjQyMzUxMTA0Mzg0NzI4NzY4NDEyODAzODU0ODMwMDQ3NzcwNDEx" \
        "MDAyODE1MzM3NDU3ODAxMDg3MjgxNTg1MzMzNDE3MTYxMTE4NzQ1MTU3MjYxOTUwNzI4NzEyNDgw" \
        "MzI4NTM1ODM1ODU3MzQyNzE0NjcyODE2NTgzNDI4NTE0NTg1MzIwMzgxODU3NDY0NzUwODI3OTQ0" \
        "ODg3NjEzMTUzNDMxMTUxNzYzNDU1NzE0MDA1MDUzNDIxODE0ODYyNDIzODM2MTczMzQ0NjAwNTIw" \
        "NTU2NDYyNDgxODYzNDA2MzA4MTE0ODUxMTQ2Mzg2MzYyMjQ4Mjc3MjIyMjUzNjMxMjI1MjEzMTU0" \
        "NjI1NjIyMjM3MTA4NjEwNjI0NTYyNTMxNTA2ODEyMjQ2MzYzNzE0MzY4MDU1MTgxNTQ2NTU3MTMx" \
        "NTI0NzU4MTU2NjAxMjY0MjA1MDU2MzcwMDM3NzcwMjA0MTYxMzE3MjQxMTI2NzYzMzUyNjY3NTQ1" \
        "NTA1MTUxNTc2NTEzMTUwNjcxNDcyMDI2MTQyMjczNTI4NzExNjA4NTU3NjIzMzMxMzU0NDM1Mzgw" \
        "MTI0MTQzMTU1MTMyNzc4ODI1MjcyMjUwMjY4MzYyMDUzMjQzNjA0MTYyMzkhB8fSAAAAAElFTkQAAAAACg=="

    assert list(ZTNR.translate(data)) == [
        (
            "HD_READY",
            "https://rtvehlsvodlote7modo2.rtve.es/mediavodv2/resources/TE_NGVA/mp4/5/3/1656573649835.mp4/video.m3u8"
            + "?hls_no_audio_only=true&idasset=6638770"
        ),
    ]



"""
$description Live TV channels and video on-demand service from NOS, a Dutch public, state-owned broadcaster.
$url nos.nl
$type live, vod
$region Netherlands
"""

import logging
import re

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:\w+\.)?nos\.nl/(?:livestream|collectie|video|uitzendingen)",
))
class NOS(Plugin):
    _msg_live_offline = "This livestream is offline."
    vod_keys = {
        "pages/Collection/Video/Video": "item",
        "pages/Video/Video": "video",
    }

    def _get_streams(self):
        try:
            scripts = self.session.http.get(self.url, schema=validate.Schema(
                validate.parse_html(),
                validate.xml_findall(".//script[@type='application/json'][@data-ssr-name]"),
                [
                    validate.union((
                        validate.get("data-ssr-name"),
                        validate.all(
                            validate.getattr("text"),
                            validate.parse_json()
                        )
                    ))
                ]
            ))
        except PluginError:
            log.error("Could not find any stream data")
            return

        for _data_ssr_name, _data_json in scripts:
            video_url = None
            log.trace(f"Found _data_ssr_name={_data_ssr_name}")

            if _data_ssr_name == "pages/Broadcasts/Broadcasts":
                self.title, video_url, is_live = validate.Schema(
                    {"currentLivestream": {
                        "is_live": bool,
                        "title": str,
                        "stream": validate.url(),
                    }},
                    validate.get("currentLivestream"),
                    validate.union_get("title", "stream", "is_live")
                ).validate(_data_json)
                if not is_live:
                    log.error(self._msg_live_offline)
                    continue

            elif _data_ssr_name == "pages/Livestream/Livestream":
                self.title, video_url, is_live = validate.Schema(
                    {
                        "streamIsLive": bool,
                        "title": str,
                        "stream": validate.url(),
                    },
                    validate.union_get("title", "stream", "streamIsLive")
                ).validate(_data_json)
                if not is_live:
                    log.error(self._msg_live_offline)
                    continue

            elif _data_ssr_name in self.vod_keys.keys():
                _key = self.vod_keys[_data_ssr_name]
                self.title, video_url = validate.Schema(
                    {_key: {
                        "title": str,
                        "aspect_ratios": {
                            "profiles": validate.all(
                                [{
                                    "name": str,
                                    "url": validate.url(),
                                }],
                                validate.filter(lambda p: p["name"] == "hls_unencrypted")
                            )
                        }
                    }},
                    validate.get(_key),
                    validate.union_get("title", ("aspect_ratios", "profiles", 0, "url"))
                ).validate(_data_json)

            if video_url is not None:
                return HLSStream.parse_variant_playlist(self.session, video_url)


__plugin__ = NOS



from streamlink.plugins.foxtr import FoxTR
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlFoxTR(PluginCanHandleUrl):
    __plugin__ = FoxTR

    should_match = [
        'http://www.fox.com.tr/canli-yayin',
    ]



"""
$description Live TV channel and video on-demand service from TV5Monde, a French free-to-air broadcaster.
$url tv5monde.com
$url tivi5mondeplus.com
$type live, vod
$region France, Belgium, Switzerland
"""

import re
from urllib.parse import urlparse

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream
from streamlink.utils.url import update_scheme


@pluginmatcher(re.compile(r"""
    https?://(?:[\w-]+\.)*(?:tv5monde|tivi5mondeplus)\.com/
""", re.VERBOSE))
class TV5Monde(Plugin):
    def _get_hls(self, root):
        schema_live = validate.Schema(
            validate.xml_xpath_string(".//*[contains(@data-broadcast,'m3u8')]/@data-broadcast"),
            str,
            validate.parse_json(),
            validate.any(
                validate.all({"files": list}, validate.get("files")),
                list
            ),
            [{
                "url": validate.url(path=validate.endswith(".m3u8"))
            }],
            validate.get((0, "url")),
            validate.transform(lambda content_url: update_scheme("https://", content_url))
        )
        try:
            live = schema_live.validate(root)
        except PluginError:
            return

        return HLSStream.parse_variant_playlist(self.session, live)

    def _get_vod(self, root):
        schema_vod = validate.Schema(
            validate.xml_xpath_string(".//script[@type='application/ld+json'][contains(text(),'VideoObject')][1]/text()"),
            str,
            validate.transform(lambda jsonlike: re.sub(r"[\r\n]+", "", jsonlike)),
            validate.parse_json(),
            validate.any(
                validate.all(
                    {"@graph": [dict]},
                    validate.get("@graph"),
                    validate.filter(lambda obj: obj["@type"] == "VideoObject"),
                    validate.get(0)
                ),
                dict
            ),
            {"contentUrl": validate.url()},
            validate.get("contentUrl"),
            validate.transform(lambda content_url: update_scheme("https://", content_url))
        )
        try:
            vod = schema_vod.validate(root)
        except PluginError:
            return

        if urlparse(vod).path.endswith(".m3u8"):
            return HLSStream.parse_variant_playlist(self.session, vod)

        return {"vod": HTTPStream(self.session, vod)}

    def _get_streams(self):
        root = self.session.http.get(self.url, schema=validate.Schema(
            validate.parse_html()
        ))

        return self._get_hls(root) or self._get_vod(root)


__plugin__ = TV5Monde



from unittest.mock import Mock

import pytest

from streamlink.plugins.youtube import YouTube
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlYouTube(PluginCanHandleUrl):
    __plugin__ = YouTube

    should_match = [
        "https://www.youtube.com/CHANNELNAME",
        "https://www.youtube.com/CHANNELNAME/",
        "https://www.youtube.com/CHANNELNAME/live",
        "https://www.youtube.com/CHANNELNAME/live/",
        "https://www.youtube.com/c/CHANNELNAME",
        "https://www.youtube.com/c/CHANNELNAME/",
        "https://www.youtube.com/c/CHANNELNAME/live",
        "https://www.youtube.com/c/CHANNELNAME/live/",
        "https://www.youtube.com/user/CHANNELNAME",
        "https://www.youtube.com/user/CHANNELNAME/",
        "https://www.youtube.com/user/CHANNELNAME/live",
        "https://www.youtube.com/user/CHANNELNAME/live/",
        "https://www.youtube.com/channel/CHANNELID",
        "https://www.youtube.com/channel/CHANNELID/",
        "https://www.youtube.com/channel/CHANNELID/live",
        "https://www.youtube.com/channel/CHANNELID/live/",
        "https://www.youtube.com/embed/aqz-KE-bpKQ",
        "https://www.youtube.com/embed/live_stream?channel=UCNye-wNBqNL5ZzHSJj3l8Bg",
        "https://www.youtube.com/v/aqz-KE-bpKQ",
        "https://www.youtube.com/watch?v=aqz-KE-bpKQ",
        "https://www.youtube.com/watch?foo=bar&baz=qux&v=aqz-KE-bpKQ",
        "https://youtu.be/0123456789A",
    ]

    should_match_groups = [
        ("https://www.youtube.com/v/aqz-KE-bpKQ", {
            "video_id": "aqz-KE-bpKQ",
        }),
        ("https://www.youtube.com/embed/aqz-KE-bpKQ", {
            "embed": "embed",
            "video_id": "aqz-KE-bpKQ",
        }),
        ("https://www.youtube.com/watch?v=aqz-KE-bpKQ", {
            "video_id": "aqz-KE-bpKQ",
        }),
    ]

    should_not_match = [
        "https://accounts.google.com/",
        "https://www.youtube.com",
        "https://www.youtube.com/",
        "https://www.youtube.com/feed/guide_builder",
        "https://www.youtube.com/t/terms",
        "https://youtu.be",
        "https://youtu.be/",
        "https://youtu.be/c/CHANNELNAME",
        "https://youtu.be/c/CHANNELNAME/live",
    ]


@pytest.mark.parametrize("url,expected", [
    ("http://gaming.youtube.com/watch?v=0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
    ("http://youtu.be/0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
    ("http://youtube.com/embed/0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
    ("http://youtube.com/embed/live_stream?channel=CHANNELID", "https://www.youtube.com/channel/CHANNELID/live"),
    ("http://www.youtube.com/watch?v=0123456789A", "https://www.youtube.com/watch?v=0123456789A"),
])
def test_translate_url(url, expected):
    assert YouTube(Mock(), url).url == expected



from streamlink.plugins.cinergroup import CinerGroup
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlCinerGroup(PluginCanHandleUrl):
    __plugin__ = CinerGroup

    should_match = [
        "https://bloomberght.com/tv",
        "https://haberturk.com/canliyayin",
        "https://haberturk.com/tv/canliyayin",
        "http://haberturk.tv/canliyayin",
        "http://showmax.com.tr/canliyayin",
        "http://showmax.com.tr/canli-yayin",
        "https://showturk.com.tr/canli-yayin/showturk",
        "https://showturk.com.tr/canli-yayin",
        "https://showturk.com.tr/canliyayin",
        "https://showtv.com.tr/canli-yayin",
        "https://showtv.com.tr/canli-yayin/showtv",
    ]



from streamlink.plugins.wasd import WASD
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlWasd(PluginCanHandleUrl):
    __plugin__ = WASD

    should_match = [
        'https://wasd.tv/channel',
        'https://wasd.tv/channel/',
    ]

    should_not_match = [
        'https://wasd.tv/channel/12345',
        'https://wasd.tv/channel/12345/videos/67890',
        'https://wasd.tv/voodik/videos?record=123456',
    ]



import argparse
import logging
import os
import platform
import re
import signal
import sys
import warnings
from contextlib import closing, suppress
from gettext import gettext
from pathlib import Path
from time import sleep
from typing import Any, Dict, List, Optional, Type, Union

import streamlink.logger as logger
from streamlink import NoPluginError, PluginError, StreamError, Streamlink, __version__ as streamlink_version
from streamlink.exceptions import FatalPluginError, StreamlinkDeprecationWarning
from streamlink.plugin import Plugin, PluginOptions
from streamlink.stream.stream import Stream, StreamIO
from streamlink.utils.named_pipe import NamedPipe
from streamlink_cli.argparser import ArgumentParser, build_parser, setup_session_options
from streamlink_cli.compat import DeprecatedPath, importlib_metadata, stdout
from streamlink_cli.console import ConsoleOutput, ConsoleUserInputRequester
from streamlink_cli.constants import CONFIG_FILES, DEFAULT_STREAM_METADATA, LOG_DIR, PLUGIN_DIRS, STREAM_SYNONYMS
from streamlink_cli.output import FileOutput, PlayerOutput
from streamlink_cli.streamrunner import StreamRunner
from streamlink_cli.utils import Formatter, HTTPServer, datetime
from streamlink_cli.utils.versioncheck import check_version


QUIET_OPTIONS = ("json", "stream_url", "quiet")


args: Any = None  # type: ignore[assignment]
console: ConsoleOutput = None  # type: ignore[assignment]
output: Union[FileOutput, PlayerOutput] = None  # type: ignore[assignment]
stream_fd: StreamIO = None  # type: ignore[assignment]
streamlink: Streamlink = None  # type: ignore[assignment]


log = logging.getLogger("streamlink.cli")


def get_formatter(plugin: Plugin):
    return Formatter(
        {
            "url": lambda: args.url,
            "plugin": lambda: plugin.module,
            "id": lambda: plugin.get_id(),
            "author": lambda: plugin.get_author(),
            "category": lambda: plugin.get_category(),
            "game": lambda: plugin.get_category(),
            "title": lambda: plugin.get_title(),
            "time": lambda: datetime.now()
        },
        {
            "time": lambda dt, fmt: dt.strftime(fmt)
        }
    )


def check_file_output(path: Path, force):
    """Checks if file already exists and ask the user if it should
    be overwritten if it does."""

    log.info(f"Writing output to\n{path.resolve()}")
    log.debug("Checking file output")

    if path.is_file() and not force:
        if sys.stdin.isatty():
            answer = console.ask(f"File {path} already exists! Overwrite it? [y/N] ")
            if not answer or answer.lower() != "y":
                sys.exit()
        else:
            log.error(f"File {path} already exists, use --force to overwrite it.")
            sys.exit()

    return FileOutput(path)


def create_output(formatter: Formatter) -> Union[FileOutput, PlayerOutput]:
    """Decides where to write the stream.

    Depending on arguments it can be one of these:
     - The stdout pipe
     - A subprocess' stdin pipe
     - A named pipe that the subprocess reads from
     - A regular file

    """

    if (args.output or args.stdout) and (args.record or args.record_and_pipe):
        console.exit("Cannot use record options with other file output options.")
        return  # type: ignore

    if args.output:
        if args.output == "-":
            return FileOutput(fd=stdout)
        else:
            return check_file_output(formatter.path(args.output, args.fs_safe_rules), args.force)

    elif args.stdout:
        return FileOutput(fd=stdout)

    elif args.record_and_pipe:
        record = check_file_output(formatter.path(args.record_and_pipe, args.fs_safe_rules), args.force)
        return FileOutput(fd=stdout, record=record)

    elif not args.player:
        console.exit(
            "The default player (VLC) does not seem to be "
            "installed. You must specify the path to a player "
            "executable with --player, a file path to save the "
            "stream with --output, or pipe the stream to "
            "another program with --stdout."
        )
        return  # type: ignore

    else:
        http = namedpipe = record = None

        if args.player_fifo:
            try:
                namedpipe = NamedPipe()  # type: ignore[abstract]  # ???
            except OSError as err:
                console.exit(f"Failed to create pipe: {err}")
                return  # type: ignore
        elif args.player_http:
            http = create_http_server()

        if args.record:
            if args.record == "-":
                record = FileOutput(fd=stdout)
            else:
                record = check_file_output(formatter.path(args.record, args.fs_safe_rules), args.force)

        log.info(f"Starting player: {args.player}")

        return PlayerOutput(
            args.player,
            args=args.player_args,
            quiet=not args.verbose_player,
            kill=not args.player_no_close,
            namedpipe=namedpipe,
            http=http,
            record=record,
            title=formatter.title(args.title, defaults=DEFAULT_STREAM_METADATA) if args.title else args.url
        )


def create_http_server(*_args, **_kwargs):
    """Creates an HTTP server listening on a given host and port.

    If host is empty, listen on all available interfaces, and if port is 0,
    listen on a random high port.
    """

    try:
        http = HTTPServer()
        http.bind(*_args, **_kwargs)
    except OSError as err:
        console.exit(f"Failed to create HTTP server: {err}")
        return

    return http


def iter_http_requests(server, player):
    """Repeatedly accept HTTP connections on a server.

    Forever if the serving externally, or while a player is running if it is not
    empty.
    """

    while not player or player.running:
        try:
            yield server.open(timeout=2.5)
        except OSError:
            continue


def output_stream_http(
    plugin: Plugin,
    initial_streams: Dict[str, Stream],
    formatter: Formatter,
    external: bool = False,
    continuous: bool = True,
    port: int = 0,
):
    """Continuously output the stream over HTTP."""
    global output

    if not external:
        if not args.player:
            console.exit("The default player (VLC) does not seem to be "
                         "installed. You must specify the path to a player "
                         "executable with --player.")

        server = create_http_server()
        player = output = PlayerOutput(
            args.player,
            args=args.player_args,
            filename=server.url,
            quiet=not args.verbose_player,
            title=formatter.title(args.title, defaults=DEFAULT_STREAM_METADATA) if args.title else args.url
        )

        try:
            log.info(f"Starting player: {args.player}")
            if player:
                player.open()
        except OSError as err:
            console.exit(f"Failed to start player: {args.player} ({err})")
    else:
        server = create_http_server(host=None, port=port)
        player = None

        log.info("Starting server, access with one of:")
        for url in server.urls:
            log.info(f" {url}")

    initial_streams_used = False
    for req in iter_http_requests(server, player):
        user_agent = req.headers.get("User-Agent") or "unknown player"
        log.info(f"Got HTTP request from {user_agent}")

        stream_fd = prebuffer = None
        while not stream_fd and (not player or player.running):
            try:
                if not initial_streams_used:
                    streams = initial_streams
                    initial_streams_used = True
                else:
                    streams = fetch_streams(plugin)

                for stream_name in (resolve_stream_name(streams, s) for s in args.stream):
                    if stream_name in streams:
                        stream = streams[stream_name]
                        break
                else:
                    log.info("Stream not available, will re-fetch streams in 10 sec")
                    sleep(10)
                    continue
            except PluginError as err:
                log.error(f"Unable to fetch new streams: {err}")
                continue

            try:
                log.info(f"Opening stream: {stream_name} ({type(stream).shortname()})")
                stream_fd, prebuffer = open_stream(stream)
            except StreamError as err:
                log.error(err)

        if stream_fd and prebuffer:
            log.debug("Writing stream to player")
            stream_runner = StreamRunner(stream_fd, server)
            try:
                stream_runner.run(prebuffer)
            except OSError as err:
                # TODO: refactor all console.exit() calls
                console.exit(str(err))

        if not continuous:
            break

        server.close(True)

    if player:
        player.close()
    server.close()


def output_stream_passthrough(stream, formatter: Formatter):
    """Prepares a filename to be passed to the player."""
    global output

    try:
        url = stream.to_url()
    except TypeError:
        console.exit("The stream specified cannot be translated to a URL")
        return False

    output = PlayerOutput(
        args.player,
        args=args.player_args,
        filename=f'"{url}"',
        call=True,
        quiet=not args.verbose_player,
        title=formatter.title(args.title, defaults=DEFAULT_STREAM_METADATA) if args.title else args.url
    )

    try:
        log.info(f"Starting player: {args.player}")
        output.open()
    except OSError as err:
        console.exit(f"Failed to start player: {args.player} ({err})")
        return False

    return True


def open_stream(stream):
    """Opens a stream and reads 8192 bytes from it.

    This is useful to check if a stream actually has data
    before opening the output.

    """
    global stream_fd

    # Attempts to open the stream
    try:
        stream_fd = stream.open()
    except StreamError as err:
        raise StreamError(f"Could not open stream: {err}")

    # Read 8192 bytes before proceeding to check for errors.
    # This is to avoid opening the output unnecessarily.
    try:
        log.debug("Pre-buffering 8192 bytes")
        prebuffer = stream_fd.read(8192)
    except OSError as err:
        stream_fd.close()
        raise StreamError(f"Failed to read data from stream: {err}")

    if not prebuffer:
        stream_fd.close()
        raise StreamError("No data returned from stream")

    return stream_fd, prebuffer


def output_stream(stream, formatter: Formatter):
    """Open stream, create output and finally write the stream to output."""
    global output

    # create output before opening the stream, so file outputs can prompt on existing output
    output = create_output(formatter)

    success_open = False
    for i in range(args.retry_open):
        try:
            stream_fd, prebuffer = open_stream(stream)
            success_open = True
            break
        except StreamError as err:
            log.error(f"Try {i + 1}/{args.retry_open}: Could not open stream {stream} ({err})")

    if not success_open:
        return console.exit(f"Could not open stream {stream}, tried {args.retry_open} times, exiting")

    try:
        output.open()
    except OSError as err:
        if isinstance(output, PlayerOutput):
            console.exit(f"Failed to start player: {args.player} ({err})")
        elif output.filename:
            console.exit(f"Failed to open output: {output.filename} ({err})")
        else:
            console.exit(f"Failed to open output ({err}")
        return

    try:
        with closing(output):
            log.debug("Writing stream to output")
            # TODO: finally clean up the global variable mess and refactor the streamlink_cli package
            # noinspection PyUnboundLocalVariable
            stream_runner = StreamRunner(stream_fd, output, args.force_progress)
            # noinspection PyUnboundLocalVariable
            stream_runner.run(prebuffer)
    except OSError as err:
        # TODO: refactor all console.exit() calls
        console.exit(str(err))

    return True


def handle_stream(plugin: Plugin, streams: Dict[str, Stream], stream_name: str) -> None:
    """Decides what to do with the selected stream.

    Depending on arguments it can be one of these:
     - Output JSON represenation
     - Output the stream URL
     - Continuously output the stream over HTTP
     - Output stream data to selected output

    """

    stream_name = resolve_stream_name(streams, stream_name)
    stream = streams[stream_name]

    # Print JSON representation of the stream
    if args.json:
        console.msg_json(
            stream,
            metadata=plugin.get_metadata()
        )

    elif args.stream_url:
        try:
            console.msg(stream.to_url())
        except TypeError:
            console.exit("The stream specified cannot be translated to a URL")

    else:
        # Find any streams with a '_alt' suffix and attempt
        # to use these in case the main stream is not usable.
        alt_streams = list(filter(lambda k: f"{stream_name}_alt" in k, sorted(streams.keys())))

        file_output = args.output or args.stdout

        formatter = get_formatter(plugin)

        for stream_name in [stream_name] + alt_streams:
            stream = streams[stream_name]
            stream_type = type(stream).shortname()

            if stream_type in args.player_passthrough and not file_output:
                log.info(f"Opening stream: {stream_name} ({stream_type})")
                success = output_stream_passthrough(stream, formatter)
            elif args.player_external_http:
                return output_stream_http(
                    plugin,
                    streams,
                    formatter,
                    external=True,
                    continuous=args.player_external_http_continuous,
                    port=args.player_external_http_port,
                )
            elif args.player_continuous_http and not file_output:
                return output_stream_http(plugin, streams, formatter)
            else:
                log.info(f"Opening stream: {stream_name} ({stream_type})")
                success = output_stream(stream, formatter)

            if success:
                break


def fetch_streams(plugin: Plugin) -> Dict[str, Stream]:
    """Fetches streams using correct parameters."""

    return plugin.streams(stream_types=args.stream_types,
                          sorting_excludes=args.stream_sorting_excludes)


def fetch_streams_with_retry(plugin: Plugin, interval: float, count: int) -> Optional[Dict[str, Stream]]:
    """Attempts to fetch streams repeatedly
       until some are returned or limit hit."""

    try:
        streams = fetch_streams(plugin)
    except PluginError as err:
        log.error(err)
        streams = None

    if not streams:
        log.info(f"Waiting for streams, retrying every {interval} second(s)")
    attempts = 0

    while not streams:
        sleep(interval)

        try:
            streams = fetch_streams(plugin)
        except FatalPluginError:
            raise
        except PluginError as err:
            log.error(err)

        if count > 0:
            attempts += 1
            if attempts >= count:
                break

    return streams


def resolve_stream_name(streams: Dict[str, Stream], stream_name: str) -> str:
    """Returns the real stream name of a synonym."""

    if stream_name in STREAM_SYNONYMS and stream_name in streams:
        for name, stream in streams.items():
            if stream is streams[stream_name] and name not in STREAM_SYNONYMS:
                return name

    return stream_name


def format_valid_streams(plugin: Plugin, streams: Dict[str, Stream]) -> str:
    """Formats a dict of streams.

    Filters out synonyms and displays them next to
    the stream they point to.

    Streams are sorted according to their quality
    (based on plugin.stream_weight).

    """

    delimiter = ", "
    validstreams = []

    for name, stream in sorted(streams.items(),
                               key=lambda stream: plugin.stream_weight(stream[0])):
        if name in STREAM_SYNONYMS:
            continue

        def synonymfilter(n):
            return stream is streams[n] and n is not name

        synonyms = list(filter(synonymfilter, streams.keys()))

        if len(synonyms) > 0:
            joined = delimiter.join(synonyms)
            name = f"{name} ({joined})"

        validstreams.append(name)

    return delimiter.join(validstreams)


def handle_url():
    """The URL handler.

    Attempts to resolve the URL to a plugin and then attempts
    to fetch a list of available streams.

    Proceeds to handle stream if user specified a valid one,
    otherwise output list of valid streams.

    """

    try:
        pluginname, pluginclass, resolved_url = streamlink.resolve_url(args.url)
        setup_plugin_options(streamlink, pluginname, pluginclass)
        plugin = pluginclass(streamlink, resolved_url)
        log.info(f"Found matching plugin {pluginname} for URL {args.url}")

        if args.retry_max or args.retry_streams:
            retry_streams = 1
            retry_max = 0
            if args.retry_streams:
                retry_streams = args.retry_streams
            if args.retry_max:
                retry_max = args.retry_max
            streams = fetch_streams_with_retry(plugin, retry_streams, retry_max)
        else:
            streams = fetch_streams(plugin)
    except NoPluginError:
        console.exit(f"No plugin can handle URL: {args.url}")
    except PluginError as err:
        console.exit(str(err))

    if not streams:
        console.exit(f"No playable streams found on this URL: {args.url}")

    if args.default_stream and not args.stream and not args.json:
        args.stream = args.default_stream

    if args.stream:
        validstreams = format_valid_streams(plugin, streams)
        for stream_name in args.stream:
            if stream_name in streams:
                log.info(f"Available streams: {validstreams}")
                handle_stream(plugin, streams, stream_name)
                return

        errmsg = f"The specified stream(s) '{', '.join(args.stream)}' could not be found"
        if args.json:
            console.msg_json(
                plugin=plugin.module,
                metadata=plugin.get_metadata(),
                streams=streams,
                error=errmsg
            )
        else:
            console.exit(f"{errmsg}.\n       Available streams: {validstreams}")
    elif args.json:
        console.msg_json(
            plugin=plugin.module,
            metadata=plugin.get_metadata(),
            streams=streams
        )
    elif args.stream_url:
        try:
            console.msg(streams[list(streams)[-1]].to_manifest_url())
        except TypeError:
            console.exit("The stream specified cannot be translated to a URL")
    else:
        validstreams = format_valid_streams(plugin, streams)
        console.msg(f"Available streams: {validstreams}")


def print_plugins():
    """Outputs a list of all plugins Streamlink has loaded."""

    pluginlist = list(streamlink.get_plugins().keys())
    pluginlist_formatted = ", ".join(sorted(pluginlist))

    if args.json:
        console.msg_json(pluginlist)
    else:
        console.msg(f"Loaded plugins: {pluginlist_formatted}")


def load_plugins(dirs: List[Path], showwarning: bool = True):
    """Attempts to load plugins from a list of directories."""
    for directory in dirs:
        if directory.is_dir():
            success = streamlink.load_plugins(str(directory))
            if success and type(directory) is DeprecatedPath:
                warnings.warn(
                    f"Loaded plugins from deprecated path, see CLI docs for how to migrate: {directory}",
                    StreamlinkDeprecationWarning,
                )
        elif showwarning:
            log.warning(f"Plugin path {directory} does not exist or is not a directory!")


def setup_args(
    parser: argparse.ArgumentParser,
    config_files: Optional[List[Path]] = None,
    ignore_unknown: bool = False,
):
    """Parses arguments."""
    global args
    arglist = sys.argv[1:]

    # Load arguments from config files
    configs = [f"@{config_file}" for config_file in config_files or []]

    args, unknown = parser.parse_known_args(configs + arglist)
    if unknown and not ignore_unknown:
        msg = gettext("unrecognized arguments: %s")
        parser.error(msg % " ".join(unknown))

    # Force lowercase to allow case-insensitive lookup
    if args.stream:
        args.stream = [stream.lower() for stream in args.stream]

    if not args.url and args.url_param:
        args.url = args.url_param


def setup_config_args(parser, ignore_unknown=False):
    config_files = []

    if args.config:
        # We want the config specified last to get the highest priority
        config_files.extend(
            config_file
            for config_file in map(lambda path: Path(path).expanduser(), reversed(args.config))
            if config_file.is_file()
        )

    else:
        # Only load first available default config
        for config_file in filter(lambda path: path.is_file(), CONFIG_FILES):  # pragma: no branch
            if type(config_file) is DeprecatedPath:
                warnings.warn(
                    f"Loaded config from deprecated path, see CLI docs for how to migrate: {config_file}",
                    StreamlinkDeprecationWarning,
                )
            config_files.append(config_file)
            break

    if streamlink and args.url:
        # Only load first available plugin config
        with suppress(NoPluginError):
            pluginname, pluginclass, resolved_url = streamlink.resolve_url(args.url)
            for config_file in CONFIG_FILES:  # pragma: no branch
                config_file = config_file.with_name(f"{config_file.name}.{pluginname}")
                if not config_file.is_file():
                    continue
                if type(config_file) is DeprecatedPath:
                    warnings.warn(
                        f"Loaded plugin config from deprecated path, see CLI docs for how to migrate: {config_file}",
                        StreamlinkDeprecationWarning,
                    )
                config_files.append(config_file)
                break

    if config_files:
        setup_args(parser, config_files, ignore_unknown=ignore_unknown)


def setup_signals():
    # restore default behavior of raising a KeyboardInterrupt on SIGINT (and SIGTERM)
    # so cleanup code can be run when the user stops execution
    signal.signal(signal.SIGINT, signal.default_int_handler)
    signal.signal(signal.SIGTERM, signal.default_int_handler)


def setup_plugins(extra_plugin_dir=None):
    """Loads any additional plugins."""
    load_plugins(PLUGIN_DIRS, showwarning=False)

    if extra_plugin_dir:
        load_plugins([Path(path).expanduser() for path in extra_plugin_dir])


def setup_streamlink():
    """Creates the Streamlink session."""
    global streamlink

    streamlink = Streamlink({"user-input-requester": ConsoleUserInputRequester(console)})


def setup_plugin_args(session: Streamlink, parser: ArgumentParser):
    """Sets Streamlink plugin options."""

    plugin_args = parser.add_argument_group("Plugin options")
    for pname, plugin in session.plugins.items():
        defaults = {}
        group = parser.add_argument_group(pname.capitalize(), parent=plugin_args)

        for parg in plugin.arguments or []:
            if not parg.is_global:
                group.add_argument(parg.argument_name(pname), **parg.options)
                defaults[parg.dest] = parg.default
            else:
                pargdest = parg.dest
                for action in parser._actions:
                    # find matching global argument
                    if pargdest != action.dest:
                        continue
                    defaults[pargdest] = action.default

                    # add plugin to global argument
                    plugins = getattr(action, "plugins", [])
                    plugins.append(pname)
                    setattr(action, "plugins", plugins)

        plugin.options = PluginOptions(defaults)


def setup_plugin_options(session: Streamlink, pluginname: str, pluginclass: Type[Plugin]):
    """Sets Streamlink plugin options."""
    if pluginclass.arguments is None:
        return

    required = {}

    for parg in pluginclass.arguments:
        if parg.options.get("help") == argparse.SUPPRESS:
            continue

        value = getattr(args, parg.dest if parg.is_global else parg.namespace_dest(pluginname))
        session.set_plugin_option(pluginname, parg.dest, value)

        if not parg.is_global:
            if parg.required:
                required[parg.name] = parg
            # if the value is set, check to see if any of the required arguments are not set
            if parg.required or value:
                try:
                    for rparg in pluginclass.arguments.requires(parg.name):
                        required[rparg.name] = rparg
                except RuntimeError:
                    log.error(f"{pluginname} plugin has a configuration error and the arguments cannot be parsed")
                    break

    if required:
        for req in required.values():
            if not session.get_plugin_option(pluginname, req.dest):
                prompt = f"{req.prompt or f'Enter {pluginname} {req.name}'}: "
                session.set_plugin_option(
                    pluginname,
                    req.dest,
                    console.askpass(prompt) if req.sensitive else console.ask(prompt)
                )


def log_root_warning():
    if hasattr(os, "geteuid"):  # pragma: no branch
        if os.geteuid() == 0:
            log.info("streamlink is running as root! Be careful!")


def log_current_versions():
    """Show current installed versions"""
    if not logger.root.isEnabledFor(logging.DEBUG):
        return

    # macOS
    if sys.platform == "darwin":
        os_version = f"macOS {platform.mac_ver()[0]}"
    # Windows
    elif sys.platform == "win32":
        os_version = f"{platform.system()} {platform.release()}"
    # Linux / other
    else:
        os_version = platform.platform()

    log.debug(f"OS:         {os_version}")
    log.debug(f"Python:     {platform.python_version()}")
    log.debug(f"Streamlink: {streamlink_version}")

    # https://peps.python.org/pep-0508/#names
    re_name = re.compile(r"[A-Z\d](?:[A-Z\d._-]*[A-Z\d])?", re.IGNORECASE)
    log.debug("Dependencies:")
    for name in [
        match.group(0)
        for match in map(re_name.match, importlib_metadata.requires("streamlink"))
        if match is not None
    ]:
        try:
            version = importlib_metadata.version(name)
        except importlib_metadata.PackageNotFoundError:
            continue
        log.debug(f" {name}: {version}")


def log_current_arguments(session: Streamlink, parser: argparse.ArgumentParser):
    global args
    if not logger.root.isEnabledFor(logging.DEBUG):
        return

    sensitive = set()
    for pname, plugin in session.plugins.items():
        for parg in plugin.arguments or []:
            if parg.sensitive:
                sensitive.add(parg.argument_name(pname))

    log.debug("Arguments:")
    for action in parser._actions:
        if not hasattr(args, action.dest):
            continue
        value = getattr(args, action.dest)
        if action.default != value:
            name = next(  # pragma: no branch
                (option for option in action.option_strings if option.startswith("--")),
                action.option_strings[0]
            ) if action.option_strings else action.dest
            log.debug(f" {name}={value if name not in sensitive else '*' * 8}")


def setup_logger_and_console(stream=sys.stdout, filename=None, level="info", json=False):
    global console

    if filename == "-":
        filename = LOG_DIR / f"{datetime.now()}.log"
    elif filename:
        filename = Path(filename).expanduser().resolve()

    if filename:
        filename.parent.mkdir(parents=True, exist_ok=True)

    verbose = level in ("trace", "all")
    streamhandler = logger.basicConfig(
        stream=stream,
        filename=filename,
        level=level,
        style="{",
        format=f"{'[{asctime}]' if verbose else ''}[{{name}}][{{levelname}}] {{message}}",
        datefmt=f"%H:%M:%S{'.%f' if verbose else ''}",
        capture_warnings=True,
    )

    console = ConsoleOutput(streamhandler.stream, json)


def main():
    error_code = 0
    parser = build_parser()

    setup_args(parser, ignore_unknown=True)
    # call argument set up as early as possible to load args from config files
    setup_config_args(parser, ignore_unknown=True)

    # Console output should be on stderr if we are outputting
    # a stream to stdout.
    if args.stdout or args.output == "-" or args.record == "-" or args.record_and_pipe:
        console_out = sys.stderr
    else:
        console_out = sys.stdout

    # We don't want log output when we are printing JSON or a command-line.
    silent_log = any(getattr(args, attr) for attr in QUIET_OPTIONS)
    log_level = args.loglevel if not silent_log else "none"
    log_file = args.logfile if log_level != "none" else None
    setup_logger_and_console(console_out, log_file, log_level, args.json)

    setup_streamlink()
    # load additional plugins
    setup_plugins(args.plugin_dirs)
    setup_plugin_args(streamlink, parser)
    # call setup args again once the plugin specific args have been added
    setup_args(parser)
    setup_config_args(parser)

    # update the logging level if changed by a plugin specific config
    log_level = args.loglevel if not silent_log else "none"
    logger.root.setLevel(log_level)

    log_root_warning()
    log_current_versions()
    log_current_arguments(streamlink, parser)

    setup_session_options(streamlink, args)

    setup_signals()

    if args.version_check or args.auto_version_check:
        try:
            check_version(force=args.version_check)
        except KeyboardInterrupt:
            error_code = 130

    if args.version_check:
        pass
    elif args.help:
        parser.print_help()
    elif args.plugins:
        print_plugins()
    elif args.can_handle_url:
        try:
            streamlink.resolve_url(args.can_handle_url)
        except NoPluginError:
            error_code = 1
        except KeyboardInterrupt:
            error_code = 130
    elif args.can_handle_url_no_redirect:
        try:
            streamlink.resolve_url_no_redirect(args.can_handle_url_no_redirect)
        except NoPluginError:
            error_code = 1
        except KeyboardInterrupt:
            error_code = 130
    elif args.url:
        try:
            handle_url()
        except KeyboardInterrupt:
            # Close output
            if output:
                output.close()
            console.msg("Interrupted! Exiting...")
            error_code = 130
        finally:
            if stream_fd:
                try:
                    log.info("Closing currently open stream...")
                    stream_fd.close()
                except KeyboardInterrupt:
                    error_code = 130
    else:
        usage = parser.format_usage()
        console.msg(
            f"{usage}\n"
            f"Use -h/--help to see the available options or read the manual at https://streamlink.github.io"
        )

    sys.exit(error_code)


def parser_helper():
    session = Streamlink()
    parser = build_parser()
    setup_plugin_args(session, parser)
    return parser



from asyncio import Future, get_event_loop
from contextlib import contextmanager
from dataclasses import dataclass
from functools import wraps
from threading import Event
from typing import Awaitable, Callable, Generator, Optional, Type


@dataclass
class _HandshakeContext:
    error: Optional[Exception] = None


class Handshake:
    """
    Control execution flow between one producer thread (application logic) and one consumer thread (tests),
    to be able to assert application state at certain points during execution.
    """

    def __init__(self) -> None:
        self._ready = Event()
        self._go = Event()
        self._done = Event()
        self._context = _HandshakeContext()

    @contextmanager
    def __call__(self, exception: Optional[Type[Exception]] = None) -> Generator[_HandshakeContext, None, None]:
        """Execute application logic in this context manager and optionally capture exceptions."""
        try:
            self.ready()
            yield self._context
        except BaseException as err:
            if exception is not None and isinstance(err, exception):
                self._context.error = err
            else:
                raise
        finally:
            self.done()

    def ready(self) -> None:
        """Make producer thread wait indefinitely until consumer thread allows one execution step."""
        self._ready.set()
        self._go.wait()
        self._go.clear()
        self._context.error = None

    def done(self):
        """Tell the cosumer thread that execution has finished."""
        self._ready.clear()
        self._done.set()

    def go(self):
        """Allow producer thread to run."""
        self._go.set()

    def wait_ready(self, timeout: Optional[float] = None) -> bool:
        """Wait for producer thread to be ready and return whether it is ready or not."""
        return self._ready.wait(timeout=timeout)

    def wait_done(self, timeout: Optional[float] = None) -> bool:
        """Wait for producer thread to be done and return whether it is done or not. If an exception was captured, raise it."""
        result = self._done.wait(timeout=timeout)
        self._done.clear()

        error = self._context.error
        if error is not None:
            self._context.error = None
            raise error

        return result

    def step(self, timeout: Optional[float] = None) -> bool:
        """Allow producer thread to run, wait for it to complete and return whether it has finished or not."""
        self.go()
        return self.wait_done(timeout=timeout)

    is_ready: Callable[[Optional[float]], Awaitable[bool]]
    is_done: Callable[[Optional[float]], Awaitable[bool]]
    asyncstep: Callable[[Optional[float]], Awaitable[bool]]


def _sync2async(obj, name, method):
    meth = getattr(obj, method)

    @wraps(meth)
    async def wrapper(*args, **kwargs):
        async def task():
            try:
                value = meth(*args, **kwargs)
            except Exception as err:
                future.set_exception(err)
            else:
                future.set_result(value)

        future = Future()
        get_event_loop().create_task(task())

        return await future

    setattr(obj, name, wrapper)


_sync2async(Handshake, "is_ready", "wait_ready")
_sync2async(Handshake, "is_done", "wait_done")
_sync2async(Handshake, "asyncstep", "step")

del _sync2async



"""
$description Live TV channels from CT, a Czech public, state-owned broadcaster.
$url ceskatelevize.cz
$type live
$region Czechia
"""

import json
import logging
import re
from urllib.parse import urlparse

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(r"https?://ct24\.ceskatelevize\.cz/"))
@pluginmatcher(re.compile(r"https?://decko\.ceskatelevize\.cz/"))
@pluginmatcher(re.compile(r"https?://sport\.ceskatelevize\.cz/"))
@pluginmatcher(re.compile(r"https?://(?:www\.)?ceskatelevize\.cz/zive/\w+"))
class Ceskatelevize(Plugin):
    schema_playlist = {
        "playlist": [{
            "streamUrls": {
                "main": validate.url(),
            },
        }],
    }

    def get_stream_url(self, video_id):
        url = self.session.http.post(
            "https://www.ceskatelevize.cz/ivysilani/ajax/get-client-playlist/",
            data={
                "playlist[0][type]": "channel",
                "playlist[0][id]": video_id,
                "requestUrl": "/ivysilani/embed/iFramePlayer.php",
                "requestSource": "iVysilani",
                "type": "html",
                "canPlayDRM": "false",
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "url": validate.url(),
                },
                validate.get("url"),
            ),
        )

        return self.session.http.get(
            url,
            schema=validate.Schema(
                validate.parse_json(),
                self.schema_playlist,
                validate.get(("playlist", 0, "streamUrls", "main")),
            ),
        )

    def get_ct24(self):
        self.id = 24
        self.title = "ČT24"
        return self.get_stream_url(self.id)

    def get_decko(self):
        self.id = 5
        self.title = "Déčko"
        return self.get_stream_url(self.id)

    def get_sport(self, content):
        video_id, key, date = validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//section[@id='live']/@data-ctcomp-data"),
            str,
            validate.parse_json(),
            {
                "items": [{
                    "items": [{
                        validate.optional("video"): {
                            "data": {
                                "source": {
                                    "playlist": [{
                                        "id": int,
                                        "key": str,
                                        "date": str,
                                        "noDrmData": {
                                            "id": int,
                                            "key": str,
                                            "drm": int,
                                            "quality": str,
                                            "assetId": str,
                                        },
                                    }],
                                },
                            },
                        },
                    }],
                }],
            },
            validate.get(("items", 0, "items", 0, "video", "data", "source", "playlist", 0)),
            validate.union_get(("noDrmData", "id"), ("noDrmData", "key"), "date"),
        ).validate(content)

        self.id = video_id
        self.title = "ČT sport"

        return self.session.http.post(
            "https://playlist.ceskatelevize.cz/",
            data={
                "data": json.dumps(
                    {
                        "contentType": "live",
                        "items": [{
                            "id": video_id,
                            "key": f"{key}",
                            "assetId": "CT4DRM",
                            "playerType": "dash",
                            "date": f"{date}",
                            "requestSource": "front-sport",
                            "drm": 0,
                            "quality": "web",
                        }],
                    },
                    separators=(',', ':'),
                ),
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "RESULT": self.schema_playlist,
                },
                validate.get(("RESULT", "playlist", 0, "streamUrls", "main")),
            ),
        )

    def get_channel(self, content):
        data = validate.Schema(
            validate.parse_html(),
            validate.xml_xpath_string(".//script[@id='__NEXT_DATA__'][text()]/text()"),
            str,
            validate.parse_json(),
            {
                "props": {
                    "pageProps": {
                        validate.optional("data"): {
                            "liveBroadcast": {
                                "id": str,
                                "current": validate.none_or_all(
                                    {
                                        "id": str,
                                        "channelName": str,
                                        "isPlayable": bool,
                                    },
                                ),
                                "next": validate.none_or_all(
                                    {
                                        "id": str,
                                        "channelName": str,
                                        "isPlayable": bool,
                                    },
                                ),
                            },
                        },
                    },
                },
            },
            validate.get(("props", "pageProps")),
        ).validate(content)

        if not data:
            return

        log.debug(f"data={data}")
        data = data.get("data").get("liveBroadcast")
        self.id = data.get("id")

        data = data.get("current") or data.get("next")
        if not data:
            return

        self.title = data.get("channelName")

        return self.get_stream_url(self.id)

    def _get_streams(self):
        res = self.session.http.get(self.url)

        if "://ct24" in res.url:
            url = self.get_ct24()
        elif "://decko" in res.url:
            url = self.get_decko()
        elif "://sport" in res.url:
            url = self.get_sport(res.content)
        else:
            url = self.get_channel(res.content)

        if url:
            res = self.session.http.head(url, allow_redirects=True)
            log.debug(f"res.url={res.url}")
            p = urlparse(res.url).path
            if not p.split("/")[-1].startswith(str(self.id)):
                log.error("This stream is not available")
                return
            else:
                return DASHStream.parse_manifest(self.session, res.url)


__plugin__ = Ceskatelevize



"""
$description Chinese live streaming platform for live video game broadcasts.
$url zhanqi.tv
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)

API_URL = "https://www.zhanqi.tv/api/static/v2.1/room/domain/{0}.json"

STATUS_ONLINE = 4
STATUS_OFFLINE = 0

_room_schema = validate.Schema(
    {
        "data": validate.any(None, {
            "status": validate.all(
                str,
                validate.transform(int)
            ),
            "videoId": str,
        })
    },
    validate.get("data")
)


@pluginmatcher(re.compile(
    r"https?://(www\.)?zhanqi\.tv/(?P<channel>[^/]+)"
))
class Zhanqitv(Plugin):
    def _get_streams(self):
        channel = self.match.group("channel")

        res = self.session.http.get(API_URL.format(channel))
        room = self.session.http.json(res, schema=_room_schema)
        if not room:
            log.info("Not a valid room url.")
            return

        if room["status"] != STATUS_ONLINE:
            log.info("Stream currently unavailable.")
            return

        url = "http://wshdl.load.cdn.zhanqi.tv/zqlive/{room[videoId]}.flv?get_url=".format(room=room)
        stream = HTTPStream(self.session, url)
        yield "live", stream

        url = "http://dlhls.cdn.zhanqi.tv/zqlive/{room[videoId]}_1024/index.m3u8?Dnion_vsnae={room[videoId]}".format(room=room)
        stream = HLSStream(self.session, url)
        yield "live", stream


__plugin__ = Zhanqitv



from streamlink.plugins.streamable import Streamable
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlStreamable(PluginCanHandleUrl):
    __plugin__ = Streamable

    should_match = [
        'https://streamable.com/example',
    ]



from unittest.mock import Mock, patch

import pytest

from streamlink.plugin import Plugin
from streamlink_cli.main import get_formatter
from streamlink_cli.utils import datetime


@pytest.fixture(scope="module")
def plugin():
    class FakePlugin(Plugin):
        __module__ = "FAKE"

        def _get_streams(self):  # pragma: no cover
            pass

    plugin = FakePlugin(Mock(), "https://foo/bar")
    plugin.id = "ID"
    plugin.author = "AUTHOR"
    plugin.category = "CATEGORY"
    plugin.title = "TITLE"

    return plugin


@pytest.mark.parametrize("formatterinput,expected", [
    ("{url}", "https://foo/bar"),
    ("{plugin}", "FAKE"),
    ("{id}", "ID"),
    ("{author}", "AUTHOR"),
    ("{category}", "CATEGORY"),
    ("{game}", "CATEGORY"),
    ("{title}", "TITLE"),
    ("{time}", "2000-01-01_00-00-00"),
    ("{time:%Y}", "2000"),
])
# workaround for freezegun not being able to patch the subclassed datetime class in streamlink_cli.utils
# which defines the default datetime->str conversion format (needed for path outputs)
@patch("streamlink_cli.utils.datetime.now", Mock(return_value=datetime(2000, 1, 1, 0, 0, 0, 0)))
@patch("streamlink_cli.main.args", Mock(url="https://foo/bar"))
def test_get_formatter(plugin, formatterinput, expected):
    formatter = get_formatter(plugin)
    assert formatter.title(formatterinput) == expected



from streamlink.plugins.bilibili import Bilibili
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlBilibili(PluginCanHandleUrl):
    __plugin__ = Bilibili

    should_match = [
        'https://live.bilibili.com/123123123',
    ]



"""
Stream wrapper around a file
"""
from streamlink.stream.stream import Stream


class FileStream(Stream):
    __shortname__ = "file"

    def __init__(self, session, path=None, fileobj=None):
        super().__init__(session)
        self.path = path
        self.fileobj = fileobj
        if not self.path and not self.fileobj:
            raise ValueError("path or fileobj must be set")

    def __json__(self):
        json = super().__json__()

        if self.path:
            json["path"] = self.path

        return json

    def to_url(self):
        if self.path is None:
            return super().to_url()

        return self.path

    def open(self):
        return self.fileobj or open(self.path)



from streamlink.plugins.dogus import Dogus
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlDogus(PluginCanHandleUrl):
    __plugin__ = Dogus

    should_match = [
        "http://eurostartv.com.tr/canli-izle",
        "https://www.kralmuzik.com.tr/tv/kral-pop-tv",
        "https://www.kralmuzik.com.tr/tv/kral-tv",
        "https://www.ntv.com.tr/canli-yayin/ntv?youtube=true",
        "https://www.startv.com.tr/canli-yayin",
    ]



from streamlink.plugins.rtbf import RTBF
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlRTBF(PluginCanHandleUrl):
    __plugin__ = RTBF

    should_match = [
        "https://www.rtbf.be/auvio/direct_doc-shot?lid=122046#/",
        "https://www.rtbf.be/auvio/emissions/detail_dans-la-toile?id=11493",
        "http://www.rtbfradioplayer.be/radio/liveradio/purefm",
    ]

    should_not_match = [
        "http://www.rtbf.be/",
        "http://www.rtbf.be/auvio",
        "http://www.rtbfradioplayer.be/",
    ]



from argparse import ArgumentParser, Namespace
from typing import Any, List
from unittest.mock import Mock

import pytest

from streamlink.session import Streamlink
from streamlink_cli.argparser import build_parser, setup_session_options
from streamlink_cli.main import main as streamlink_cli_main


@pytest.fixture(scope="module")
def parser():
    return build_parser()


@pytest.fixture
def session(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink.session.Streamlink.load_builtin_plugins", lambda _: None)
    yield Streamlink()


@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize("argv,option,expected", [
    pytest.param(
        ["--locale", "xx_XX"],
        "locale",
        "xx_XX",
        id="Arg+value without mapper",
    ),
    pytest.param(
        ["--http-query-param", "foo=bar", "--http-query-param", "baz=qux"],
        "http-query-params",
        {"foo": "bar", "baz": "qux"},
        id="Arg+value with dict mapper",
    ),
    pytest.param(
        ["--http-no-ssl-verify"],
        "http-ssl-verify",
        False,
        id="Arg with bool mapper",
    ),
    pytest.param(
        ["--http-ssl-cert-crt-key", "foo.crt", "bar.key"],
        "http-ssl-cert",
        ("foo.crt", "bar.key"),
        id="Arg+value with tuple mapper",
    ),
    pytest.param(
        ["--hls-timeout", "123"],
        "stream-timeout",
        123.0,
        id="Deprecated argument",
    ),
    pytest.param(
        ["--hls-timeout", "123", "--stream-timeout", "456"],
        "stream-timeout",
        456.0,
        id="Deprecated argument with override",
    ),
])
def test_setup_session_options(parser: ArgumentParser, session: Streamlink, argv: List, option: str, expected: Any):
    args = parser.parse_args(argv)
    setup_session_options(session, args)
    assert session.get_option(option) == expected


def test_cli_main_setup_session_options(monkeypatch: pytest.MonkeyPatch, parser: ArgumentParser, session: Streamlink):
    class StopTest(Exception):
        pass

    mock_setup_session_options = Mock()

    monkeypatch.setattr("sys.argv", [])
    monkeypatch.setattr("streamlink_cli.main.CONFIG_FILES", [])
    monkeypatch.setattr("streamlink_cli.main.logger", Mock())
    monkeypatch.setattr("streamlink_cli.main.streamlink", session)
    monkeypatch.setattr("streamlink_cli.main.build_parser", Mock(return_value=parser))
    monkeypatch.setattr("streamlink_cli.main.setup_streamlink", Mock())
    monkeypatch.setattr("streamlink_cli.main.setup_plugins", Mock())
    monkeypatch.setattr("streamlink_cli.main.log_root_warning", Mock())
    monkeypatch.setattr("streamlink_cli.main.log_current_versions", Mock())
    monkeypatch.setattr("streamlink_cli.main.log_current_arguments", Mock())
    monkeypatch.setattr("streamlink_cli.main.setup_session_options", mock_setup_session_options)
    monkeypatch.setattr("streamlink_cli.main.setup_signals", Mock(side_effect=StopTest))

    with pytest.raises(StopTest):
        streamlink_cli_main()

    assert mock_setup_session_options.call_count == 1, \
        "Has called setup_session_options() before setting up signals and running actual CLI code"
    assert mock_setup_session_options.call_args_list[0][0][0] is session
    assert isinstance(mock_setup_session_options.call_args_list[0][0][1], Namespace)



from unittest.mock import Mock, call, patch

import pytest

from streamlink.exceptions import FatalPluginError
from streamlink.session import Streamlink
from streamlink_cli.console import ConsoleUserInputRequester
from tests.plugin.testplugin import TestPlugin as _TestPlugin


def test_session():
    console_input = ConsoleUserInputRequester(Mock())
    session = Streamlink({"user-input-requester": console_input})
    assert session.get_option("user-input-requester") is console_input


class TestPluginUserInput:
    @pytest.fixture
    def session(self):
        return Streamlink()

    @pytest.fixture
    def testplugin(self, session: Streamlink):
        return _TestPlugin(session, "http://example.com/stream")

    @pytest.fixture
    def console_input(self, request, session: Streamlink):
        isatty: bool = request.param
        with patch("streamlink_cli.console.sys.stdin.isatty", return_value=isatty):
            mock_console = Mock()
            mock_console.ask.return_value = "username"
            mock_console.askpass.return_value = "password"
            console_input = ConsoleUserInputRequester(mock_console)
            session.set_option("user-input-requester", console_input)
            yield console_input

    def test_user_input_not_implemented(self, testplugin: _TestPlugin):
        with pytest.raises(FatalPluginError) as cm:
            testplugin.input_ask("test")
        assert str(cm.value) == "This plugin requires user input, however it is not supported on this platform"

        with pytest.raises(FatalPluginError) as cm:
            testplugin.input_ask_password("test")
        assert str(cm.value) == "This plugin requires user input, however it is not supported on this platform"

    @pytest.mark.parametrize("console_input", [True], indirect=True)
    def test_user_input_console(self, testplugin: _TestPlugin, console_input: ConsoleUserInputRequester):
        assert testplugin.input_ask("username") == "username"
        assert console_input.console.ask.call_args_list == [call("username: ")]

        assert testplugin.input_ask_password("password") == "password"
        assert console_input.console.askpass.call_args_list == [call("password: ")]

    @pytest.mark.parametrize("console_input", [False], indirect=True)
    def test_user_input_console_no_tty(self, testplugin: _TestPlugin, console_input: ConsoleUserInputRequester):
        with pytest.raises(FatalPluginError) as cm:
            testplugin.input_ask("username")
        assert str(cm.value) == "User input error: no TTY available"

        with pytest.raises(FatalPluginError) as cm:
            testplugin.input_ask_password("username")
        assert str(cm.value) == "User input error: no TTY available"



from streamlink.plugins.tv8 import TV8
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlTV8(PluginCanHandleUrl):
    __plugin__ = TV8

    should_match = [
        'https://www.tv8.com.tr/canli-yayin',
    ]



import unittest
from io import StringIO
from unittest.mock import Mock, patch

from streamlink_cli.console import ConsoleOutput


class TestConsoleOutput(unittest.TestCase):
    def test_msg(self):
        output = StringIO()
        console = ConsoleOutput(output)
        console.msg("foo")
        console.msg_json({"test": 1})
        self.assertEqual("foo\n", output.getvalue())

    def test_msg_json(self):
        output = StringIO()
        console = ConsoleOutput(output, json=True)
        console.msg("foo")
        console.msg_json({"test": 1})
        self.assertEqual('{\n  "test": 1\n}\n', output.getvalue())

    def test_msg_json_object(self):
        output = StringIO()
        console = ConsoleOutput(output, json=True)
        console.msg_json(Mock(__json__=Mock(return_value={"test": 1})))
        self.assertEqual('{\n  "test": 1\n}\n', output.getvalue())

    def test_msg_json_list(self):
        output = StringIO()
        console = ConsoleOutput(output, json=True)
        test_list = ["foo", "bar"]
        console.msg_json(test_list)
        self.assertEqual('[\n  "foo",\n  "bar"\n]\n', output.getvalue())

    def test_msg_json_merge_object(self):
        output = StringIO()
        console = ConsoleOutput(output, json=True)
        test_obj1 = {"test": 1, "foo": "foo"}
        test_obj2 = Mock(__json__=Mock(return_value={"test": 2}))
        console.msg_json(test_obj1, test_obj2, ["qux"], foo="bar", baz="qux")
        self.assertEqual(
            '{\n'
            '  "test": 2,\n'
            '  "foo": "bar",\n'
            '  "baz": "qux"\n'
            '}\n',
            output.getvalue()
        )
        self.assertEqual([("test", 1), ("foo", "foo")], list(test_obj1.items()))

    def test_msg_json_merge_list(self):
        output = StringIO()
        console = ConsoleOutput(output, json=True)
        test_list1 = ["foo", "bar"]
        test_list2 = Mock(__json__=Mock(return_value={"foo": "bar"}))
        console.msg_json(test_list1, ["baz"], test_list2, {"foo": "bar"}, foo="bar", baz="qux")
        self.assertEqual(
            '[\n'
            '  "foo",\n'
            '  "bar",\n'
            '  "baz",\n'
            '  {\n    "foo": "bar"\n  },\n'
            '  {\n    "foo": "bar"\n  },\n'
            '  {\n    "foo": "bar",\n    "baz": "qux"\n  }\n'
            ']\n',
            output.getvalue()
        )
        self.assertEqual(["foo", "bar"], test_list1)

    @patch("streamlink_cli.console.sys.exit")
    def test_msg_json_error(self, mock_exit):
        output = StringIO()
        console = ConsoleOutput(output, json=True)
        console.msg_json({"error": "bad"})
        self.assertEqual('{\n  "error": "bad"\n}\n', output.getvalue())
        mock_exit.assert_called_with(1)

    @patch("streamlink_cli.console.sys.exit")
    def test_exit(self, mock_exit: Mock):
        output = StringIO()
        console = ConsoleOutput(output)
        console.exit("error")
        self.assertEqual("error: error\n", output.getvalue())
        mock_exit.assert_called_with(1)

    @patch("streamlink_cli.console.sys.exit")
    def test_exit_json(self, mock_exit: Mock):
        output = StringIO()
        console = ConsoleOutput(output, json=True)
        console.exit("error")
        self.assertEqual('{\n  "error": "error"\n}\n', output.getvalue())
        mock_exit.assert_called_with(1)

    @patch("streamlink_cli.console.input", Mock(return_value="hello"))
    @patch("streamlink_cli.console.sys.stdin.isatty", Mock(return_value=True))
    def test_ask(self):
        output = StringIO()
        console = ConsoleOutput(output)
        self.assertEqual("hello", console.ask("test: "))
        self.assertEqual("test: ", output.getvalue())

    @patch("streamlink_cli.console.input")
    @patch("streamlink_cli.console.sys.stdin.isatty", Mock(return_value=False))
    def test_ask_no_tty(self, mock_input: Mock):
        output = StringIO()
        console = ConsoleOutput(output)
        self.assertIsNone(console.ask("test: "))
        self.assertEqual("", output.getvalue())
        mock_input.assert_not_called()

    @patch("streamlink_cli.console.input", Mock(side_effect=ValueError))
    @patch("streamlink_cli.console.sys.stdin.isatty", Mock(return_value=True))
    def test_ask_input_exception(self):
        output = StringIO()
        console = ConsoleOutput(output)
        self.assertIsNone(console.ask("test: "))
        self.assertEqual("test: ", output.getvalue())

    @patch("streamlink_cli.console.getpass")
    @patch("streamlink_cli.console.sys.stdin.isatty", Mock(return_value=True))
    def test_askpass(self, mock_getpass: Mock):
        def getpass(prompt, stream):
            stream.write(prompt)
            return "hello"

        output = StringIO()
        console = ConsoleOutput(output)
        mock_getpass.side_effect = getpass
        self.assertEqual("hello", console.askpass("test: "))
        self.assertEqual("test: ", output.getvalue())

    @patch("streamlink_cli.console.sys.stdin.isatty", Mock(return_value=False))
    def test_askpass_no_tty(self):
        output = StringIO()
        console = ConsoleOutput(output)
        self.assertIsNone(console.askpass("test: "))



"""
$description French live TV channel and video on-demand service owned by Gulli.
$url replay.gulli.fr
$type live, vod
$region France
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r'https?://replay\.gulli\.fr/(?:Direct|.+/(?P<video_id>VOD\d+))'
))
class Gulli(Plugin):
    LIVE_PLAYER_URL = 'https://replay.gulli.fr/jwplayer/embedstreamtv'
    VOD_PLAYER_URL = 'https://replay.gulli.fr/jwplayer/embed/{0}'

    _playlist_re = re.compile(r'sources: (\[.+?\])', re.DOTALL)
    _vod_video_index_re = re.compile(r'jwplayer\(idplayer\).playlistItem\((?P<video_index>[0-9]+)\)')
    _mp4_bitrate_re = re.compile(r'.*_(?P<bitrate>[0-9]+)\.mp4')

    _video_schema = validate.Schema(
        validate.all(
            validate.transform(lambda x: re.sub(r'"?file"?:\s*[\'"](.+?)[\'"],?', r'"file": "\1"', x, flags=re.DOTALL)),
            validate.transform(lambda x: re.sub(r'"?\w+?"?:\s*function\b.*?(?<={).*(?=})', "", x, flags=re.DOTALL)),
            validate.parse_json(),
            [
                validate.Schema({
                    'file': validate.url()
                })
            ]
        )
    )

    def _get_streams(self):
        video_id = self.match.group('video_id')
        if video_id is not None:
            # VOD
            live = False
            player_url = self.VOD_PLAYER_URL.format(video_id)
        else:
            # Live
            live = True
            player_url = self.LIVE_PLAYER_URL

        res = self.session.http.get(player_url)
        playlist = re.findall(self._playlist_re, res.text)
        index = 0
        if not live:
            # Get the index for the video on the playlist
            match = self._vod_video_index_re.search(res.text)
            if match is None:
                return
            index = int(match.group('video_index'))

        if not playlist:
            return
        videos = self._video_schema.validate(playlist[index])

        for video in videos:
            video_url = video['file']

            # Ignore non-supported MSS streams
            if 'isml/Manifest' in video_url:
                continue

            try:
                if '.m3u8' in video_url:
                    yield from HLSStream.parse_variant_playlist(self.session, video_url).items()
                elif '.mp4' in video_url:
                    match = self._mp4_bitrate_re.match(video_url)
                    bitrate = "vod" if match is None else f"{match.group('bitrate')}k"
                    yield bitrate, HTTPStream(self.session, video_url)
            except OSError as err:
                if '403 Client Error' in str(err):
                    log.error('Failed to access stream, may be due to geo-restriction')
                raise


__plugin__ = Gulli



from streamlink.plugins.invintus import InvintusMedia
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlInvintusMedia(PluginCanHandleUrl):
    __plugin__ = InvintusMedia

    should_match = [
        'https://player.invintus.com/?clientID=9375922947&eventID=2020031185',
        'https://player.invintus.com/?clientID=9375922947&eventID=2020031184',
        'https://player.invintus.com/?clientID=9375922947&eventID=2020031183',
        'https://player.invintus.com/?clientID=9375922947&eventID=2020031182',
        'https://player.invintus.com/?clientID=9375922947&eventID=2020031181'
    ]



import os
import signal

import freezegun.config
import pytest

# import streamlink_cli as early as possible to execute its default signal overrides
# noinspection PyUnresolvedReferences
import streamlink_cli  # noqa: F401


# immediately restore default signal handlers for the test runner
signal.signal(signal.SIGINT, signal.default_int_handler)
signal.signal(signal.SIGTERM, signal.default_int_handler)


# make freezegun ignore the following pytest modules, so it doesn't mess with pytest's --duration=N report
# when freezing time in fixtures, even though ["_pytest.runner.", "_pytest.terminal."] is already included
# in freezegun's default module ignore list (notice the trailing dots).
freezegun.config.configure(extend_ignore_list=["_pytest.runner", "_pytest.terminal"])


# make pytest rewrite assertions in dynamically parametrized plugin tests
# https://docs.pytest.org/en/stable/how-to/writing_plugins.html#assertion-rewriting
pytest.register_assert_rewrite("tests.plugins")


windows_only = pytest.mark.skipif(os.name != "nt", reason="test only applicable on Windows")
posix_only = pytest.mark.skipif(os.name != "posix", reason="test only applicable on a POSIX OS")


__all__ = ["windows_only", "posix_only"]



"""
$description Russian live-streaming platform for gaming and esports, owned by VKontakte.
$url vkplay.live
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://vkplay\.live/(?P<channel_name>\w+)/?$",
))
class VKplay(Plugin):
    API_URL = "https://api.vkplay.live/v1"

    def _get_streams(self):
        self.author = self.match.group("channel_name")
        log.debug(f"Channel name: {self.author}")

        data = self.session.http.get(
            f"{self.API_URL}/blog/{self.author}/public_video_stream",
            headers={"Referer": self.url},
            acceptable_status=(200, 404),
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {"error": str, "error_description": str},
                        validate.get("error_description"),
                    ),
                    validate.all(
                        {
                            "category": {
                                "title": str,
                            },
                            "title": str,
                            "data": validate.any(
                                [
                                    validate.all(
                                        {
                                            "vid": str,
                                            "playerUrls": [
                                                validate.all(
                                                    {
                                                        "type": str,
                                                        "url": validate.any("", validate.url()),
                                                    },
                                                    validate.union_get("type", "url"),
                                                ),
                                            ],
                                        },
                                        validate.union_get("vid", "playerUrls"),
                                    ),
                                ],
                                [],
                            ),
                        },
                        validate.union_get(
                            ("category", "title"),
                            "title",
                            ("data", 0),
                        ),
                    ),
                ),
            ),
        )
        if type(data) is str:
            log.error(data)
            return

        self.category, self.title, streamdata = data
        if not streamdata:
            return

        self.id, streams = streamdata

        for streamtype, streamurl in streams:
            if streamurl and streamtype == "live_hls":
                return HLSStream.parse_variant_playlist(self.session, streamurl)


__plugin__ = VKplay



#!/usr/bin/env python3
import os
import sys

from streamlink import __version__ as streamlink_version

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('.'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '3.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.intersphinx',
    'ext_html_template_vars',
    'ext_argparse',
    'ext_github',
    'ext_plugins',
    'ext_releaseref',
    'myst_parser',
]

autosectionlabel_prefix_document = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Streamlink'
copyright = '2023, Streamlink'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

#
# The short X.Y version.
version = streamlink_version.split('+')[0]
# The full version, including alpha/beta/rc tags.
release = streamlink_version

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '_applications.rst']

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
#pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

github_project = 'streamlink/streamlink'


# -- Options for autodoc ---------------------------------------------------

autodoc_default_options = {
    "show-inheritance": True,
    "members": True,
    "member-order": "groupwise",  # autodoc_member_order
    "class-doc-from": "both",  # autoclass_content
}
autodoc_inherit_docstrings = False
autodoc_typehints = "description"


# -- Options for intersphinx ---------------------------------------------------

intersphinx_mapping = {
    # "python": ("https://docs.python.org/3", None),
    "requests": ("https://requests.readthedocs.io/en/stable/", None),
}

intersphinx_timeout = 60


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'furo'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "source_repository": "https://github.com/streamlink/streamlink/",
    "source_branch": "master",
    "source_directory": "docs/",
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# HTML template variables via the custom ext_html_template_vars extension
html_template_vars = {
    "oneliner": (
        "A command-line utility that extracts streams from various services and pipes them into a video player of choice."
    ),
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "../icon.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = [
    'styles/custom.css',
    'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.1/css/fontawesome.min.css',
    'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.1/css/solid.min.css',
    'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.1/css/brands.min.css',
]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    '**': [
        'sidebar/scroll-start.html',
        'sidebar/brand.html',
        'sidebar/search.html',
        'sidebar/navigation.html',
        'sidebar/github-buttons.html',
        'sidebar/scroll-end.html',
    ]
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = False

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'streamlinkdoc'

# -- Options for manual page output --------------------------------------------

# Only include the man page in builds with the "man" tag set: via `-t man` (see Makefile)

# noinspection PyUnresolvedReferences
if not tags.tags.get("man"):  # type: ignore[name-defined]
    exclude_patterns.append("_man.rst")

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        "_man",
        "streamlink",
        "extracts streams from various services and pipes them into a video player of choice",
        ["Streamlink Contributors"],
        1,
    ),
]

# If true, show URL addresses after external links.
#man_show_urls = False

# If true, make a section directory on build man page.
# Always set this to false to fix inconsistencies between recent sphinx releases
man_make_section_directory = False



import re
from pathlib import Path
from typing import Callable, Optional, Union

from streamlink.compat import is_win32


REPLACEMENT = "_"
SPECIAL_PATH_PARTS = (".", "..")

_UNPRINTABLE = "".join(chr(c) for c in range(32))
_UNSUPPORTED_POSIX = "/"
_UNSUPPORTED_WIN32 = "\x7f\"*/:<>?\\|"

RE_CHARS_POSIX = re.compile(f"[{re.escape(_UNPRINTABLE + _UNSUPPORTED_POSIX)}]+")
RE_CHARS_WIN32 = re.compile(f"[{re.escape(_UNPRINTABLE + _UNSUPPORTED_WIN32)}]+")
if is_win32:
    RE_CHARS = RE_CHARS_WIN32
else:
    RE_CHARS = RE_CHARS_POSIX


def replace_chars(path: str, charmap: Optional[str] = None, replacement: str = REPLACEMENT) -> str:
    if charmap is None:
        pattern = RE_CHARS
    else:
        charmap = charmap.lower()
        if charmap in ("posix", "unix"):
            pattern = RE_CHARS_POSIX
        elif charmap in ("windows", "win32"):
            pattern = RE_CHARS_WIN32
        else:
            raise ValueError("Invalid charmap")

    return pattern.sub(replacement, path)


def replace_path(pathlike: Union[str, Path], mapper: Callable[[str], str]) -> Path:
    def get_part(part):
        newpart = mapper(part)
        return REPLACEMENT if part != newpart and newpart in SPECIAL_PATH_PARTS else newpart

    return Path(*(get_part(part) for part in Path(pathlike).expanduser().parts))



import unittest
from typing import List
from unittest.mock import ANY, MagicMock, Mock, call, patch

import pytest

from streamlink import PluginError
from streamlink.stream.dash import DASHStream, DASHStreamWorker
from streamlink.stream.dash_manifest import MPD
from tests.resources import text, xml


class TestDASHStream(unittest.TestCase):
    def setUp(self):
        self.session = MagicMock()
        self.test_url = "http://test.bar/foo.mpd"
        self.session.http.get.return_value = Mock(url=self.test_url)

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_video_only(self, mpdClass):
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=720),
                         Mock(id=2, mimeType="video/mp4", height=1080)
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p", "1080p"])
        )

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_audio_only(self, mpdClass):
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="audio/mp4", bandwidth=128.0, lang='en'),
                         Mock(id=2, mimeType="audio/mp4", bandwidth=256.0, lang='en')
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["a128k", "a256k"])
        )

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_audio_single(self, mpdClass):
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=720),
                         Mock(id=2, mimeType="video/mp4", height=1080),
                         Mock(id=3, mimeType="audio/aac", bandwidth=128.0, lang='en')
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p", "1080p"])
        )

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_audio_multi(self, mpdClass):
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=720),
                         Mock(id=2, mimeType="video/mp4", height=1080),
                         Mock(id=3, mimeType="audio/aac", bandwidth=128.0, lang='en'),
                         Mock(id=4, mimeType="audio/aac", bandwidth=256.0, lang='en')
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p+a128k", "1080p+a128k", "720p+a256k", "1080p+a256k"])
        )

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_audio_multi_lang(self, mpdClass):
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=720),
                         Mock(id=2, mimeType="video/mp4", height=1080),
                         Mock(id=3, mimeType="audio/aac", bandwidth=128.0, lang='en'),
                         Mock(id=4, mimeType="audio/aac", bandwidth=128.0, lang='es')
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p", "1080p"])
        )

        self.assertEqual(streams["720p"].audio_representation.lang, "en")
        self.assertEqual(streams["1080p"].audio_representation.lang, "en")

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_audio_multi_lang_alpha3(self, mpdClass):
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=720),
                         Mock(id=2, mimeType="video/mp4", height=1080),
                         Mock(id=3, mimeType="audio/aac", bandwidth=128.0, lang='eng'),
                         Mock(id=4, mimeType="audio/aac", bandwidth=128.0, lang='spa')
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p", "1080p"])
        )

        self.assertEqual(streams["720p"].audio_representation.lang, "eng")
        self.assertEqual(streams["1080p"].audio_representation.lang, "eng")

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_audio_invalid_lang(self, mpdClass):
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=720),
                         Mock(id=2, mimeType="video/mp4", height=1080),
                         Mock(id=3, mimeType="audio/aac", bandwidth=128.0, lang='en_no_voice'),
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p", "1080p"])
        )

        self.assertEqual(streams["720p"].audio_representation.lang, "en_no_voice")
        self.assertEqual(streams["1080p"].audio_representation.lang, "en_no_voice")

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_audio_multi_lang_locale(self, mpdClass):
        self.session.localization.language.alpha2 = "es"
        self.session.localization.explicit = True

        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=720),
                         Mock(id=2, mimeType="video/mp4", height=1080),
                         Mock(id=3, mimeType="audio/aac", bandwidth=128.0, lang='en'),
                         Mock(id=4, mimeType="audio/aac", bandwidth=128.0, lang='es')
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p", "1080p"])
        )

        self.assertEqual(streams["720p"].audio_representation.lang, "es")
        self.assertEqual(streams["1080p"].audio_representation.lang, "es")

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_drm(self, mpdClass):
        mpdClass.return_value = Mock(periods=[Mock(adaptationSets=[Mock(contentProtection="DRM")])])

        self.assertRaises(PluginError,
                          DASHStream.parse_manifest,
                          self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

    def test_parse_manifest_string(self):
        with text("dash/test_9.mpd") as mpd_txt:
            test_manifest = mpd_txt.read()

        streams = DASHStream.parse_manifest(self.session, test_manifest)
        self.assertSequenceEqual(list(streams.keys()), ['2500k'])

    @patch('streamlink.stream.dash.DASHStreamReader')
    @patch('streamlink.stream.dash.FFMPEGMuxer')
    def test_stream_open_video_only(self, muxer, reader):
        stream = DASHStream(self.session, Mock(), Mock(id=1, mimeType="video/mp4"))
        open_reader = reader.return_value = Mock()

        stream.open()

        reader.assert_called_with(stream, 1, "video/mp4")
        open_reader.open.assert_called_with()
        muxer.assert_not_called()

    @patch('streamlink.stream.dash.DASHStreamReader')
    @patch('streamlink.stream.dash.FFMPEGMuxer')
    def test_stream_open_video_audio(self, muxer, reader):
        stream = DASHStream(self.session, Mock(), Mock(id=1, mimeType="video/mp4"), Mock(id=2, mimeType="audio/mp3", lang='en'))
        open_reader = reader.return_value = Mock()

        stream.open()

        self.assertSequenceEqual(reader.mock_calls, [call(stream, 1, "video/mp4"),
                                                     call().open(),
                                                     call(stream, 2, "audio/mp3"),
                                                     call().open()])
        self.assertSequenceEqual(muxer.mock_calls, [call(self.session, open_reader, open_reader, copyts=True),
                                                    call().open()])

    @patch('streamlink.stream.dash.MPD')
    def test_segments_number_time(self, mpdClass):
        with xml("dash/test_9.mpd") as mpd_xml:
            mpdClass.return_value = MPD(mpd_xml, base_url="http://test.bar", url="http://test.bar/foo.mpd")

            streams = DASHStream.parse_manifest(self.session, self.test_url)
            mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

            self.assertSequenceEqual(list(streams.keys()), ['2500k'])

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_with_duplicated_resolutions(self, mpdClass):
        """
            Verify the fix for https://github.com/streamlink/streamlink/issues/3365
        """
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=1080, bandwidth=128.0),
                         Mock(id=2, mimeType="video/mp4", height=1080, bandwidth=64.0),
                         Mock(id=3, mimeType="video/mp4", height=1080, bandwidth=32.0),
                         Mock(id=4, mimeType="video/mp4", height=720),
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertSequenceEqual(
            sorted(list(streams.keys())),
            sorted(["720p", "1080p", "1080p_alt", "1080p_alt2"])
        )

    @patch('streamlink.stream.dash.MPD')
    def test_parse_manifest_with_duplicated_resolutions_sorted_bandwidth(self, mpdClass):
        """
            Verify the fix for https://github.com/streamlink/streamlink/issues/4217
        """
        mpdClass.return_value = Mock(periods=[
            Mock(adaptationSets=[
                Mock(contentProtection=None,
                     representations=[
                         Mock(id=1, mimeType="video/mp4", height=1080, bandwidth=64.0),
                         Mock(id=2, mimeType="video/mp4", height=1080, bandwidth=128.0),
                         Mock(id=3, mimeType="video/mp4", height=1080, bandwidth=32.0),
                     ])
            ])
        ])

        streams = DASHStream.parse_manifest(self.session, self.test_url)
        mpdClass.assert_called_with(ANY, base_url="http://test.bar", url="http://test.bar/foo.mpd")

        self.assertEqual(streams["1080p"].video_representation.bandwidth, 128.0)
        self.assertEqual(streams["1080p_alt"].video_representation.bandwidth, 64.0)
        self.assertEqual(streams["1080p_alt2"].video_representation.bandwidth, 32.0)


class TestDASHStreamWorker:
    @pytest.fixture
    def mock_time(self, monkeypatch: pytest.MonkeyPatch) -> Mock:
        mock = Mock(return_value=1)
        monkeypatch.setattr("streamlink.stream.dash.time", mock)
        return mock

    @pytest.fixture(autouse=True)
    def mock_wait(self, monkeypatch: pytest.MonkeyPatch) -> Mock:
        mock = Mock(return_value=True)
        monkeypatch.setattr("streamlink.stream.dash.DASHStreamWorker.wait", mock)
        return mock

    @pytest.fixture
    def representation(self) -> Mock:
        return Mock(id=1, mimeType="video/mp4", height=720)

    @pytest.fixture
    def segments(self) -> List[Mock]:
        return [
            Mock(url="init_segment"),
            Mock(url="first_segment"),
            Mock(url="second_segment"),
        ]

    @pytest.fixture
    def mpd(self, representation) -> Mock:
        return Mock(
            publishTime=1,
            minimumUpdatePeriod=Mock(total_seconds=Mock(return_value=0)),
            periods=[
                Mock(
                    duration=Mock(total_seconds=Mock(return_value=0)),
                    adaptationSets=[
                        Mock(
                            contentProtection=None,
                            representations=[representation],
                        ),
                    ],
                ),
            ],
        )

    @pytest.fixture
    def worker(self, mpd):
        reader = MagicMock(representation_id=1, mime_type="video/mp4")
        worker = DASHStreamWorker(reader)
        worker.mpd = mpd
        return worker

    def test_dynamic_reload(
        self,
        monkeypatch: pytest.MonkeyPatch,
        worker: DASHStreamWorker,
        representation: Mock,
        segments: List[Mock],
        mpd: Mock,
    ):
        mpd.dynamic = True
        mpd.type = "dynamic"
        monkeypatch.setattr("streamlink.stream.dash.MPD", lambda *args, **kwargs: mpd)

        segment_iter = worker.iter_segments()

        representation.segments.return_value = segments[:1]
        assert next(segment_iter) is segments[0]
        assert representation.segments.call_args_list == [call(init=True)]
        assert not worker._wait.is_set()

        representation.segments.reset_mock()
        representation.segments.return_value = segments[1:]
        assert [next(segment_iter), next(segment_iter)] == segments[1:]
        assert representation.segments.call_args_list == [call(), call(init=False)]
        assert not worker._wait.is_set()

    def test_static(
        self,
        worker: DASHStreamWorker,
        representation: Mock,
        segments: List[Mock],
        mpd: Mock,
    ):
        mpd.dynamic = False
        mpd.type = "static"

        representation.segments.return_value = segments
        assert list(worker.iter_segments()) == segments
        assert representation.segments.call_args_list == [call(init=True)]
        assert worker._wait.is_set()

    @pytest.mark.parametrize("duration", [
        0,
        204.32,
    ])
    def test_static_refresh_wait(
        self,
        duration: float,
        mock_wait: Mock,
        mock_time: Mock,
        worker: DASHStreamWorker,
        representation: Mock,
        segments: List[Mock],
        mpd: Mock,
    ):
        """
            Verify the fix for https://github.com/streamlink/streamlink/issues/2873
        """
        mpd.dynamic = False
        mpd.type = "static"
        mpd.periods[0].duration.total_seconds.return_value = duration

        representation.segments.return_value = segments
        assert list(worker.iter_segments()) == segments
        assert representation.segments.call_args_list == [call(init=True)]
        assert mock_wait.call_args_list == [call(5)]
        assert worker._wait.is_set()

    def test_duplicate_rep_id(self):
        representation_vid = Mock(id=1, mimeType="video/mp4", height=720)
        representation_aud = Mock(id=1, mimeType="audio/aac", lang="en")

        mpd = Mock(
            dynamic=False,
            publishTime=1,
            periods=[
                Mock(
                    adaptationSets=[
                        Mock(
                            contentProtection=None,
                            representations=[representation_vid],
                        ),
                        Mock(
                            contentProtection=None,
                            representations=[representation_aud],
                        ),
                    ],
                ),
            ],
        )

        assert DASHStreamWorker.get_representation(mpd, 1, "video/mp4") is representation_vid
        assert DASHStreamWorker.get_representation(mpd, 1, "audio/aac") is representation_aud



from streamlink.plugins.ard_mediathek import ARDMediathek
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlARDMediathek(PluginCanHandleUrl):
    __plugin__ = ARDMediathek

    should_match_groups = [
        (
            "https://www.ardmediathek.de/live/"
            + "Y3JpZDovL2Rhc2Vyc3RlLmRlL2xpdmUvY2xpcC9hYmNhMDdhMy0zNDc2LTQ4NTEtYjE2Mi1mZGU4ZjY0NmQ0YzQ",
            {"id_live": "Y3JpZDovL2Rhc2Vyc3RlLmRlL2xpdmUvY2xpcC9hYmNhMDdhMy0zNDc2LTQ4NTEtYjE2Mi1mZGU4ZjY0NmQ0YzQ"},
        ),
        (
            "https://www.ardmediathek.de/live/"
            + "Y3JpZDovL2Rhc2Vyc3RlLmRlL2xpdmUvY2xpcC9hYmNhMDdhMy0zNDc2LTQ4NTEtYjE2Mi1mZGU4ZjY0NmQ0YzQ?toolbarType=default",
            {"id_live": "Y3JpZDovL2Rhc2Vyc3RlLmRlL2xpdmUvY2xpcC9hYmNhMDdhMy0zNDc2LTQ4NTEtYjE2Mi1mZGU4ZjY0NmQ0YzQ"},
        ),
        (
            "https://www.ardmediathek.de/live/tagesschau24/"
            + "Y3JpZDovL2Rhc2Vyc3RlLmRlL3RhZ2Vzc2NoYXUvbGl2ZXN0cmVhbQ",
            {"id_live": "Y3JpZDovL2Rhc2Vyc3RlLmRlL3RhZ2Vzc2NoYXUvbGl2ZXN0cmVhbQ"},
        ),
        (
            "https://www.ardmediathek.de/video/"
            + "Y3JpZDovL2Rhc2Vyc3RlLmRlL3RhZ2Vzc2NoYXUvOWE4NGIzODgtZDEzNS00ZWU0LWI4ODEtZDYyNTQzYjg3ZmJlLzE",
            {"id_video": "Y3JpZDovL2Rhc2Vyc3RlLmRlL3RhZ2Vzc2NoYXUvOWE4NGIzODgtZDEzNS00ZWU0LWI4ODEtZDYyNTQzYjg3ZmJlLzE"},
        ),
        (
            "https://www.ardmediathek.de/video/arte/blackfish-der-killerwal/arte/"
            + "Y3JpZDovL2FydGUudHYvdmlkZW9zLzA1MDMyNy0wMDAtQQ",
            {"id_video": "Y3JpZDovL2FydGUudHYvdmlkZW9zLzA1MDMyNy0wMDAtQQ"},
        ),
        (
            "https://www.ardmediathek.de/video/expeditionen-ins-tierreich/die-revolte-der-schimpansen/ndr/"
            + "Y3JpZDovL25kci5kZS9jY2E3M2MzZS00ZTljLTRhOWItODE3MC05MjhjM2MwNWEyMDM?toolbarType=default",
            {"id_video": "Y3JpZDovL25kci5kZS9jY2E3M2MzZS00ZTljLTRhOWItODE3MC05MjhjM2MwNWEyMDM"},
        ),
    ]



"""
$description South Korean live-streaming social platform with a focus on celebrities.
$url vlive.tv
$type live
$notes Embedded Naver VODs are not supported
"""

import logging
import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://www\.vlive\.tv/(?P<format>video|post)/(?P<id>[\d-]+)"
))
class Vlive(Plugin):
    _playinfo_url = "https://www.vlive.tv/globalv-web/vam-web/old/v3/live/{0}/playInfo"

    def _get_streams(self):
        self.session.http.headers.update({"Referer": self.url})
        video_json = self.session.http.get(self.url, schema=validate.Schema(
            re.compile(r"window\.__PRELOADED_STATE__\s*=\s*({.*})\s*(?:<|,\s*function)", re.DOTALL),
            validate.none_or_all(
                validate.get(1),
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {"postDetail": {"post": {"officialVideo": {
                            "type": str,
                            "videoSeq": int,
                            validate.optional("status"): str,
                        }}}},
                        validate.get(("postDetail", "post", "officialVideo")),
                    ),
                    validate.all(
                        {"postDetail": {"error": {"errorCode": str}}},
                        validate.get(("postDetail", "error")),
                    ),
                ),
            ),
        ))
        if video_json is None:
            log.error('Could not parse video page')
            return

        err = video_json.get('errorCode')
        if err == 'common_700':
            log.error('Available only to members of the channel')
            return
        elif err == 'common_702':
            log.error('Vlive+ VODs are not supported')
            return
        elif err == 'common_404':
            log.error('Could not find video page')
            return
        elif err is not None:
            log.error('Unknown error code: {0}'.format(err))
            return

        if video_json['type'] == 'VOD':
            log.error('VODs are not supported')
            return

        url_format, video_id = self.match.groups()
        if url_format == 'post':
            video_id = str(video_json['videoSeq'])

        video_status = video_json.get('status')
        if video_status == 'ENDED':
            log.error('Stream has ended')
            return
        elif video_status != 'ON_AIR':
            log.error('Unknown video status: {0}'.format(video_status))
            return

        stream_url = self.session.http.get(
            self._playinfo_url.format(video_id),
            schema=validate.Schema(
                validate.parse_json(),
                {"result": {"adaptiveStreamUrl": validate.url()}},
                validate.get(("result", "adaptiveStreamUrl")),
            ),
        )
        return HLSStream.parse_variant_playlist(self.session, stream_url).items()


__plugin__ = Vlive



"""
$description Japanese live radio simulcasts and time-shifted broadcasts for over 100 stations.
$url radiko.jp
$type live, vod
$region Japan
"""

import base64
import datetime
import hashlib
import random
import re
from urllib.parse import urlencode

from lxml.etree import XML

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(
    r'https?://radiko\.jp/(#!/)?(?P<state>live|ts)/(?P<station_id>[a-zA-Z0-9-]+)/?(?P<start_at>\d+)?'
))
class Radiko(Plugin):
    _api_auth_1 = 'https://radiko.jp/v2/api/auth1'
    _api_auth_2 = 'https://radiko.jp/v2/api/auth2'
    _auth_key = 'bcd151073c03b352e1ef2fd66c32209da9ca0afa'

    def _get_streams(self):
        state = self.match.group('state')
        station_id = self.match.group('station_id').upper()
        if state == 'live':
            url, token = self._live(station_id)
        else:
            start_at = self.match.group('start_at')
            url, token = self._timefree(station_id, start_at)
        headers = {
            'X-Radiko-AuthToken': token
        }
        self.session.http.headers = headers
        yield from HLSStream.parse_variant_playlist(self.session, url).items()

    def _live(self, station_id):
        live_url = 'http://f-radiko.smartstream.ne.jp/{}/_definst_/simul-stream.stream/playlist.m3u8'.format(station_id)
        token, area_id = self._authorize()
        lsid = hashlib.md5(str(random.random()).encode('utf-8')).hexdigest()
        live_params = {
            'station_id': station_id,
            'l': 15,
            'lsid': lsid,
            'type': 'b'
        }
        url = f"{live_url}?{urlencode(live_params)}"
        return url, token

    def _timefree(self, station_id, start_at):
        m3u8_url = 'https://tf-rpaa.smartstream.ne.jp/tf/playlist.m3u8'
        token, area_id = self._authorize()
        lsid = hashlib.md5(str(random.random()).encode('utf-8')).hexdigest()
        end_at = self._get_xml(start_at, station_id)
        m3u8_params = {
            'station_id': station_id,
            'start_at': start_at,
            'ft': start_at,
            'end_at': end_at,
            'to': end_at,
            'l': 15,
            'lsid': lsid,
            'type': 'b'
        }
        url = f"{m3u8_url}?{urlencode(m3u8_params)}"
        return url, token

    def _authorize(self):
        headers = {
            'x-radiko-app': 'pc_html5',
            'x-radiko-app-version': '0.0.1',
            'x-radiko-device': 'pc',
            'x-radiko-user': 'dummy_user'
        }
        self.session.http.headers.update(headers)
        r = self.session.http.get(self._api_auth_1)
        token = r.headers.get('x-radiko-authtoken')
        offset = int(r.headers.get('x-radiko-keyoffset'))
        length = int(r.headers.get('x-radiko-keylength'))
        partial_key = base64.b64encode(self._auth_key[offset:offset + length].encode('ascii')).decode('utf-8')
        headers = {
            'x-radiko-authtoken': token,
            'x-radiko-device': 'pc',
            'x-radiko-partialkey': partial_key,
            'x-radiko-user': 'dummy_user'
        }
        self.session.http.headers.update(headers)
        r = self.session.http.get(self._api_auth_2)
        if r.status_code == 200:
            return token, r.text.split(',')[0]

    def _get_xml(self, start_at, station_id):
        today = datetime.date(int(start_at[:4]), int(start_at[4:6]), int(start_at[6:8]))
        yesterday = today - datetime.timedelta(days=1)
        if int(start_at[8:10]) < 5:
            date = yesterday.strftime('%Y%m%d')
        else:
            date = today.strftime('%Y%m%d')
        api = 'http://radiko.jp/v3/program/station/date/{}/{}.xml'.format(date, station_id)
        r = self.session.http.get(api)
        tree = XML(r.content)
        for x in tree[2][0][1].findall('prog'):
            if x.attrib['ft'] == start_at:
                return x.attrib['to']


__plugin__ = Radiko



from datetime import datetime
from unittest.mock import Mock, patch

import pytest
from freezegun import freeze_time

from streamlink.utils.formatter import Formatter


class TestFormatter:
    class Obj:
        def __str__(self):
            return "obj"

    @pytest.fixture
    def prop(self):
        return Mock(return_value="prop")

    @pytest.fixture
    def obj(self):
        return self.Obj()

    @pytest.fixture
    def formatter(self, prop: Mock, obj: Obj):
        with freeze_time("2000-01-02T03:04:05.000006Z"):
            yield Formatter(
                {
                    "prop": prop,
                    "obj": lambda: obj,
                    "time": datetime.now,
                    "empty": lambda: "",
                    "none": lambda: None,
                },
                {
                    "time": lambda dt, fmt: dt.strftime(fmt),
                },
            )

    def test_unknown(self, formatter: Formatter):
        assert formatter.format("{}") == "{}"
        assert formatter.format("some {unknown} variable") == "some {unknown} variable"
        assert formatter.format("some {unknown} variable", {"unknown": "known"}) == "some known variable"
        assert formatter.cache == {}

    def test_format(self, formatter: Formatter, prop: Mock, obj: Obj):
        assert formatter.format("text '{prop}' '{empty}' '{none}'") == "text 'prop' '' ''"
        assert formatter.cache == dict(prop="prop", empty="", none=None)
        assert prop.call_count == 1

        assert formatter.format("text '{prop}' '{obj}' '{empty}' '{none}'") == "text 'prop' 'obj' '' ''"
        assert formatter.cache == dict(prop="prop", obj=obj, empty="", none=None)
        assert prop.call_count == 1

        defaults = dict(prop="PROP", obj="OBJ", empty="EMPTY", none="NONE")
        assert formatter.format("'{prop}' '{obj}' '{empty}' '{none}'", defaults) == "'prop' 'obj' '' 'NONE'"
        assert formatter.cache == dict(prop="prop", obj=obj, empty="", none=None)
        assert prop.call_count == 1

    def test_format_spec(self, formatter: Formatter):
        assert formatter.format("{time}") == "2000-01-02 03:04:05.000006"
        assert formatter.cache == dict(time=datetime(2000, 1, 2, 3, 4, 5, 6, None))
        assert formatter.format("{time:%Y}") == "2000"
        assert formatter.format("{time:%Y-%m-%d}") == "2000-01-02"
        assert formatter.format("{time:%H:%M:%S}") == "03:04:05"
        with patch("datetime.datetime.strftime", side_effect=ValueError):
            assert formatter.format("{time:foo:bar}") == "{time:foo:bar}"
        assert formatter.cache == dict(time=datetime(2000, 1, 2, 3, 4, 5, 6, None))

        assert formatter.format("{prop:foo}") == "prop"
        assert formatter.format("{none:foo}") == ""
        assert formatter.format("{unknown:format}") == "{unknown:format}"
        assert formatter.format("{unknown:format}", {"unknown": "known"}) == "known"



from streamlink.plugins.mediaklikk import Mediaklikk
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMediaklikk(PluginCanHandleUrl):
    __plugin__ = Mediaklikk

    should_match = [
        'https://www.mediaklikk.hu/duna-world-elo/',
        'https://www.mediaklikk.hu/duna-world-radio-elo',
        'https://www.mediaklikk.hu/m1-elo',
        'https://www.mediaklikk.hu/m2-elo',
        'https://mediaklikk.hu/video/hirado-2021-06-24-i-adas-6/',
        'https://m4sport.hu/elo/',
        'https://m4sport.hu/elo/?channelId=m4sport+',
        'https://m4sport.hu/elo/?showchannel=mtv4plus',
        'https://m4sport.hu/euro2020-video/goool2-13-resz',
        'https://hirado.hu/videok/',
        'https://hirado.hu/videok/nemzeti-sporthirado-2021-06-24-i-adas-2/',
        'https://petofilive.hu/video/2021/06/23/the-anahit-klip-limited-edition/',
    ]



from streamlink.plugins.dailymotion import DailyMotion
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlDailyMotion(PluginCanHandleUrl):
    __plugin__ = DailyMotion

    should_match_groups = [
        ("https://www.dailymotion.com/france24", {"user": "france24"}),
        ("https://www.dailymotion.com/video/x8dmdzz", {"media_id": "x8dmdzz"}),
        ("https://www.dailymotion.com/embed/video/x8dmdzz", {"media_id": "x8dmdzz"}),
    ]

    should_not_match = [
        "https://www.dailymotion.com/",
    ]



"""
$description French television news network, covering world news from the European perspective.
$url euronews.com
$type live
"""

import logging
import re

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r"https?://(?:(?P<subdomain>\w+)\.)?euronews\.com/(?P<live>live$)?"
))
class Euronews(Plugin):
    API_URL = "https://{subdomain}.euronews.com/api/live/data"

    def _get_live(self):
        if not self.match["live"] or not self.match["subdomain"]:
            return

        try:
            log.debug("Querying live API")
            stream_url = self.session.http.get(
                self.API_URL.format(subdomain=self.match["subdomain"]),
                params={"locale": self.match["subdomain"]},
                schema=validate.Schema(
                    validate.parse_json(),
                    {"videoPrimaryUrl": validate.url(path=validate.endswith(".m3u8"))},
                    validate.get("videoPrimaryUrl"),
                ),
            )
        except PluginError:
            pass
        else:
            return HLSStream.parse_variant_playlist(self.session, stream_url)

    def _get_embed(self, root):
        schema_video_id = validate.Schema(
            validate.xml_xpath_string(".//div[@data-video][1]/@data-video"),
            str,
            validate.parse_json(),
            {
                "player": "pfp",
                "videoId": str,
            },
            validate.get("videoId"),
        )
        try:
            log.debug("Looking for YouTube video ID")
            video_id = schema_video_id.validate(root)
        except PluginError:
            pass
        else:
            return self.session.streams(f"https://www.youtube.com/watch?v={video_id}")

        schema_video_url = validate.Schema(
            validate.xml_xpath_string(".//iframe[@id='pfpPlayer'][starts-with(@src,'https://www.youtube.com/')][1]/@src"),
            str,
        )
        try:
            log.debug("Looking for embedded YouTube iframe")
            video_url = schema_video_url.validate(root)
        except PluginError:
            pass
        else:
            return self.session.streams(video_url)

    def _get_vod(self, root):
        schema_vod = validate.Schema(
            validate.any(
                validate.all(
                    validate.xml_xpath_string(".//meta[@property='og:video'][1]/@content"),
                    str,
                ),
                validate.all(
                    validate.xml_xpath_string(".//div[@data-video][1]/@data-video"),
                    str,
                    validate.parse_json(),
                    {"url": str},
                    validate.get("url"),
                ),
            ),
            validate.url(),
        )
        try:
            log.debug("Looking for VOD URL")
            video_url = schema_vod.validate(root)
        except PluginError:
            pass
        else:
            return dict(vod=HTTPStream(self.session, video_url))

    def _get_streams(self):
        live = self._get_live()
        if live:
            return live

        root = self.session.http.get(self.url, schema=validate.Schema(
            validate.parse_html(),
        ))

        return self._get_embed(root) or self._get_vod(root)


__plugin__ = Euronews



"""
$description Live TV channels from STV, a Scottish free-to-air broadcaster.
$url player.stv.tv
$type live
$region United Kingdom
"""

import logging
import re

from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.stream.hls import HLSStream

log = logging.getLogger(__name__)


@pluginmatcher(re.compile(
    r'https?://player\.stv\.tv/live'
))
class STV(Plugin):
    API_URL = 'https://player.api.stv.tv/v1/streams/stv/'

    def get_title(self):
        if self.title is None:
            self._get_api_results()
        return self.title

    def _get_api_results(self):
        res = self.session.http.get(self.API_URL)
        data = self.session.http.json(res)

        if data['success'] is False:
            raise PluginError(data['reason']['message'])

        try:
            self.title = data['results']['now']['title']
        except KeyError:
            self.title = 'STV'

        return data

    def _get_streams(self):
        hls_url = self._get_api_results()['results']['streamUrl']
        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = STV
